# 四、相同的对象

在编程世界中，重复代码被认为是邪恶的。 我们不应在不同的地方拥有相同或相似代码的多个副本。

有很多方法可以合并具有相似功能的代码段或对象。 在本章中，我们将介绍最著名的面向对象原理：继承。 正如第 1 章，“面向对象设计”中讨论的那样，继承使我们可以创建*，即两个或多个类之间的*关系，将通用逻辑抽象为超类并进行管理 子类中的特定详细信息。 特别是，我们将介绍以下方面的 Python 语法和原理：

*   基本继承
*   从内置继承
*   多重继承
*   多态性和鸭子打字

# 基本继承

从技术上讲，我们创建的每个类都使用继承。 所有 Python 类都是名为`object`的特殊类的子类。 此类提供的数据和行为很少（它提供的行为都是仅供内部使用的双下划线方法），但是它确实允许 Python 以相同的方式处理所有对象。

如果我们不显式继承其他类，则我们的类将自动继承`object`。 但是，我们可以公开声明我们的类是使用以下语法从`object`派生的：

```py
class MySubClass(object):
    pass
```

这就是继承！ 从技术上讲，该示例与第 2 章和 *Python* 中的对象的第一个示例没有什么不同，因为如果我们未明确提供[ 不同的超类。 超类或父类是从其继承的类。 子类是从超类继承的类。 在这种情况下，超类是`object`，而`MySubClass`是子类。 子类也被称为是从其父类派生的，或者该子类扩展了父类。

正如您从示例中发现的一样，继承比基本类定义所需的语法更少。 只需在类名之后但在冒号终止类定义之前的括号内包括父类的名称。 这就是我们要告诉 Python 新类应该从给定的超类派生的全部操作。

我们如何在实践中应用继承？ 继承最简单，最明显的用途是向现有类添加功能。 让我们从一个简单的联系人管理器开始，该管理器跟踪几个人的姓名和电子邮件地址。 联系人类负责维护类变量中所有联系人的列表，并初始化单个联系人的姓名和地址：

```py
class Contact:
    all_contacts = []

    def __init__(self, name, email):
        self.name = name
        self.email = email
        Contact.all_contacts.append(self)

```

这个例子向我们介绍了类变量。 由于`all_contacts`列表是类定义的一部分，因此该类的所有实例都共享它。 这意味着只有一个`Contact.all_contacts`列表，我们可以将其作为`Contact.all_contacts`进行访问。 不太明显的是，我们还可以在从`Contact`实例化的任何对象上以`self.all_contacts`的形式访问它。 如果在对象上找不到该字段，那么它将在类上找到，因此引用相同的单个列表。

### 注意

请谨慎使用此语法，因为如果*使用`self.all_contacts`设置了*变量，则实际上将创建仅与该对象关联的**新的**实例变量。 class 变量将保持不变，并可以通过`Contact.all_contacts`进行访问。

这是一个简单的类，它使我们可以跟踪有关每个联系人的几个数据。 但是，如果我们的某些联系人也是我们需要从其订购耗材的供应商怎么办？ 我们可以在`Contact`类中添加`order`方法，但这将使人们意外地从客户或家人朋友的联系人那里订购商品。 相反，让我们创建一个新的`Supplier`类，其作用类似于我们的`Contact`类，但是具有一个附加的`order`方法：

```py
class Supplier(Contact):
    def order(self, order):
        print("If this were a real system we would send "
                "'{}' order to '{}'".format(order, self.name))
```

现在，如果我们在可靠的解释器中测试此类，我们将看到所有联系人（包括供应商）都在其`__init__`中接受姓名和电子邮件地址，但只有供应商具有功能上的订购方法：

```py
>>> c = Contact("Some Body", "somebody@example.net")
>>> s = Supplier("Sup Plier", "supplier@example.net")
>>> print(c.name, c.email, s.name, s.email)
Some Body somebody@example.net Sup Plier supplier@example.net
>>> c.all_contacts
[<__main__.Contact object at 0xb7375ecc>,
 <__main__.Supplier object at 0xb7375f8c>]
>>> c.order("I need pliers")
Traceback (most recent call last):
 File "<stdin>", line 1, in <module>
AttributeError: 'Contact' object has no attribute 'order'
>>> s.order("I need pliers")
If this were a real system we would send 'I need pliers' order to
'Sup Plier '

```

因此，现在我们的`Supplier`类可以完成联系人可以做的所有事情（包括将自己添加到`all_contacts`列表中）以及作为供应商需要处理的所有特殊事项。 这就是继承之美。

## 扩展内置

这种继承的一种有趣的用法是向内置类添加功能。 在前面看到的`Contact`类中，我们将联系人添加到所有联系人的列表中。 如果我们还想按名称搜索该列表怎么办？ 好吧，我们可以在`Contact`类上添加一个方法来搜索它，但是感觉这个方法实际上属于列表本身。 我们可以使用继承来做到这一点：

```py
class ContactList(list):
    def search(self, name):
        '''Return all contacts that contain the search value
        in their name.'''
        matching_contacts = []
        for contact in self:
            if name in contact.name:
                matching_contacts.append(contact)
        return matching_contacts

class Contact:
    all_contacts = ContactList()

    def __init__(self, name, email):
        self.name = name
        self.email = email
        self.all_contacts.append(self)
```

我们没有实例化普通列表作为我们的类变量，而是创建了一个扩展内置`list`的新`ContactList`类。 然后，我们将此子类实例化为`all_contacts`列表。 我们可以如下测试新的搜索功能：

```py
>>> c1 = Contact("John A", "johna@example.net")
>>> c2 = Contact("John B", "johnb@example.net")
>>> c3 = Contact("Jenna C", "jennac@example.net")
>>> [c.name for c in Contact.all_contacts.search('John')]
['John A', 'John B']

```

您是否想知道我们如何将内置语法`[]`更改为可以继承的语法？ 实际上，使用`[]`创建一个空列表是使用`list()`创建一个空列表的简写； 这两种语法的行为相同：

```py
>>> [] == list()
True

```

实际上，`[]`语法实际上是所谓的**语法糖**，它在后台调用了`list()`构造函数。 `list`数据类型是我们可以扩展的类。 实际上，列表本身扩展了`object`类：

```py
>>> isinstance([], object)
True

```

作为第二个示例，我们可以扩展`dict`类，该类与列表类似，是使用`{}`语法简写构造的类：

```py
class LongNameDict(dict):
    def longest_key(self):
        longest = None
        for key in self:
            if not longest or len(key) > len(longest):
                longest = key
        return longest
```

这在交互式解释器中很容易测试：

```py
>>> longkeys = LongNameDict()
>>> longkeys['hello'] = 1
>>> longkeys['longest yet'] = 5
>>> longkeys['hello2'] = 'world'
>>> longkeys.longest_key()
'longest yet'

```

大多数内置类型都可以类似地扩展。 通常扩展的内置是`object`，`list`，`set`，`dict`，`file`和`str`。 有时也会继承诸如`int`和`float`之类的数字类型。

## 覆盖和超级

因此，对于*来说，继承是一个不错的，它将*新行为添加到现有类中，但是*会改变*行为呢？ 我们的`contact`类仅允许使用名称和电子邮件地址。 对于大多数联系人而言，这可能就足够了，但是如果我们想为我们的密友添加电话号码怎么办？

正如我们在第 2 章，Python 中的*对象中所看到的那样，我们可以通过在构造接触后在接触上设置一个`phone`属性来轻松地做到这一点。 但是，如果要使第三个变量在初始化时可用，则必须重写`__init__`。 覆盖表示更改或用子类中的新方法（具有相同名称）替换超类的方法。 无需特殊语法即可执行此操作； 子类的新创建的方法会自动调用，而不是超类的方法。 例如：*

```py
class Friend(Contact):
    def __init__(self, name, email, phone):
        self.name = name
        self.email = email
        self.phone = phone
```

可以覆盖的任何方法，而不仅仅是`__init__`。 但是，在继续之前，我们需要在此示例中解决一些问题。 我们的`Contact`和`Friend`类具有重复的代码来设置`name`和`email`属性； 这可能会使代码维护变得复杂，因为我们必须在两个或多个位置更新代码。 更令人震惊的是，我们的`Friend`类忽略了将自身添加到我们在`Contact`类上创建的`all_contacts`列表中。

我们真正需要的是一种在`Contact`类上执行原始`__init__`方法的方法。 这就是`super`功能的作用； 它返回该对象作为父类的实例，从而使我们可以直接调用父方法：

```py
class Friend(Contact):
    def __init__(self, name, email, phone):
        super().__init__(name, email)
        self.phone = phone
```

本示例首先使用`super`获取父对象的实例，然后在该对象上调用`__init__`，并传入期望的参数。 然后，它执行自己的初始化，即设置`phone`属性。

### 注意

请注意，`super()`语法在旧版本的 Python 中不起作用。 像列表和字典的[]和{}语法一样，它是更复杂的构造的简写形式。 在讨论多重继承时，我们将在短期内了解更多信息，但是现在知道在 Python 2 中，您必须调用`super(EmailContact, self).__init__()`。 特别注意，第一个参数是子类的名称，而不是某些人希望的作为要调用的父类的名称。 另外，请记住类在对象之前。 我总是忘了顺序，因此 Python 3 中的新语法为我节省了很多时间来查找它。

可以在任何方法内进行`super()`调用，而不仅仅是`__init__`。 这意味着可以通过覆盖和调用`super`来修改所有方法。 也可以在方法的任何点处调用`super`。 我们不必将调用作为方法的第一行。 例如，我们可能需要在将传入参数转发给超类之前对其进行操作或验证。

# 多重继承

多重继承是敏感的主题。 从原则上讲，这非常简单：从多个父类继承的子类能够从它们两个中访问功能。 实际上，这没有听起来那么有用，许多专家程序员建议不要使用它。

### 注意

根据经验，如果您认为需要多重继承，那么您可能错了，但是如果您知道需要继承，那么您可能是对的。

最简单且最有用的多重继承形式称为 **mixin**。 mixin 通常是一个超类，它本身并不存在，但可以被其他一些类继承以提供额外的功能。 例如，假设我们要向`Contact`类添加功能，该功能允许向`self.email`发送电子邮件。 发送电子邮件是我们可能想在其他许多类上使用的一项常见任务。 因此，我们可以编写一个简单的 mixin 类来为我们发送电子邮件：

```py
class MailSender:
    def send_mail(self, message):
        print("Sending mail to " + self.email)
        # Add e-mail logic here
```

为简便起见，我们将不在此处包括实际的电子邮件逻辑； 如果您对研究如何完成感兴趣，请参阅 Python 标准库中的`smtplib`模块。

该类没有做任何特殊的事情（实际上，它几乎不能作为一个独立的类起作用），但是它确实允许我们使用多重继承定义一个描述`Contact`和`MailSender`的新类：

```py
class EmailableContact(Contact, MailSender):
    pass
```

多重继承的语法看起来像类定义中的参数列表。 我们没有在括号中包含一个基类，而是包含了两个（或多个），并以逗号分隔。 我们可以测试这个新的混合动力以查看混合工作：

```py
>>> e = EmailableContact("John Smith", "jsmith@example.net")
>>> Contact.all_contacts
[<__main__.EmailableContact object at 0xb7205fac>]
>>> e.send_mail("Hello, test e-mail here")
Sending mail to jsmith@example.net

```

`Contact`初始化程序为仍将新联系人添加到`all_contacts`列表中，并且 mixin 能够向`self.email`发送邮件，因此我们知道一切正常。

这并不难，您可能想知道关于多重继承的可怕警告是什么。 我们将在一分钟内介绍复杂性，但让我们考虑一下我们拥有的其他选项，而不是在此处使用 mixin：

*   我们本可以使用单一继承并将`send_mail`函数添加到子类中。 此处的缺点是，对于所有其他需要电子邮件的类，电子邮件功能必须重复。
*   我们可以创建一个独立的 Python 函数来发送电子邮件，并在需要发送电子邮件时使用提供的正确电子邮件地址作为参数来调用该函数。
*   我们本来可以探索使用组合而不是继承的几种方法。 例如，`EmailableContact`可以具有`MailSender`对象，而不是从其继承。
*   我们可以进行猴子补丁（在第 7 章和 *Python 面向对象的快捷方式*中简要介绍猴子补丁），在`Contact`类之后添加`send_mail`方法 类已创建。 这是通过定义一个接受`self`参数的函数并将其设置为现有类的属性来完成的。

当混合来自不同类的方法时，多重继承可以正常工作，但是当我们必须在超类上调用方法时，它会变得非常混乱。 有多个超类。 我们怎么知道该叫哪一个？ 我们怎么知道用什么顺序打电话给他们？

让我们通过在`Friend`类中添加家庭住址来探讨这些问题。 我们可能会采用几种方法。 地址是代表联系人的街道，城市，国家和其他相关详细信息的字符串的集合。 我们可以将每个字符串作为参数传递给`Friend`类的`__init__`方法。 我们还可以将这些字符串存储在元组或字典中，并将它们作为单个参数传递到`__init__`中。 如果不需要在地址中添加任何方法，这可能是最好的做法。

另一个选择是创建一个新的`Address`类以将这些字符串保持在一起，然后将该类的实例传递到`Friend`类的`__init__`方法中。 此解决方案的优点是我们可以向数据添加行为（例如，提供指导或打印地图的方法），而不仅仅是静态地存储它。 正如我们在第 1 章，“面向对象设计”中讨论的那样，这是组成的示例。 组成的“具有”关系是解决此问题的完美可行的解决方案，它使我们可以在建筑物，企业或组织等其他实体中重用`Address`类。

但是，继承也是可行的解决方案，这就是我们要探讨的内容。 让我们添加一个包含地址的新类。 我们将这个新类称为“ AddressHolder”，而不是“ Address”，因为继承定义了*是*关系。 说“朋友”是“地址”是不正确的，但是由于朋友可以有“地址”，因此我们可以说“朋友”是“ AddressHolder”。 以后，我们可以创建其他拥有地址的实体（公司，建筑物）。 这是我们的`AddressHolder`类：

```py
class AddressHolder:
    def __init__(self, street, city, state, code):
        self.street = street
        self.city = city
        self.state = state
        self.code = code
```

很简单; 我们只是获取所有数据，并在初始化时将其扔到实例变量中。

## 钻石问题

我们可以使用多重继承将该新类添加为现有`Friend`类的父类。 棘手的是，我们现在有两个父级`__init__`方法，这两个方法都需要初始化。 并且它们需要使用不同的参数进行初始化。 我们如何做到这一点？ 好吧，我们可以从一个幼稚的方法开始：

```py
class Friend(Contact, AddressHolder):
    def __init__(
        self, name, email, phone,street, city, state, code):
        Contact.__init__(self, name, email)
        AddressHolder.__init__(self, street, city, state, code)
        self.phone = phone
```

在此示例中，我们直接在每个超类上调用`__init__`函数，并显式传递`self`参数。 这个例子在技术上是可行的。 我们可以直接在类上访问不同的变量。 但是有一些问题。

首先，如果我们忽略显式调用初始化器，则超类可能未初始化为。 这不会破坏本示例，但在常见情况下可能导致难以调试的程序崩溃。 想象一下，尝试将数据插入尚未连接的数据库中。

其次，更危险的是，由于类层次结构的组织，可能会多次调用超类。 看一下这个继承图：

![The diamond problem](img/8781OS_3_01.jpg)

`Friend`类中的`__init__`方法首先调用`Contact`上的`__init__`，这将隐式初始化`object`超类（请记住，所有类均源自`object`）。 `Friend`然后在`AddressHolder`上调用`__init__`，这又隐式地初始化`object`超类。 这意味着父类已设置两次。 对于`object`类，这是相对无害的，但是在某些情况下，它可能会带来灾难。 想象一下，对于每个请求尝试两次连接到数据库！

基类只能调用一次。 曾经，是的，但是什么时候？ 我们先叫`Friend`，然后叫`Contact`，然后叫`Object`，然后叫`AddressHolder`？ 还是`Friend`，然后是`Contact`，然后是`AddressHolder`，然后是`Object`？

### 注意

通过修改类的 `__mro__`（**方法解析顺序**）属性，可以即时调整方法的调用顺序。 这超出了本模块的范围。 如果您认为需要了解它，我建议 *Expert Python 编程*，*TarekZiadé*，*Packt Publishing* 或阅读有关该主题的原始文档，网址为 [] http://www.python.org/download/releases/2.3/mro/](http://www.python.org/download/releases/2.3/mro/) 。

让我们看第二个人为的例子，它更清楚地说明了这个问题。 在这里，我们有一个基类，它具有一个名为`call_me`的方法。 两个子类重写该方法，然后另一个子类使用多重继承扩展这两个方法。 由于类图的菱形形状，这被称为菱形继承：

![The diamond problem](img/8781OS_3_02.jpg)

让我们将此图转换为代码； 此示例显示了何时调用方法：

```py
class BaseClass:
    num_base_calls = 0
    def call_me(self):
        print("Calling method on Base Class")
        self.num_base_calls += 1

class LeftSubclass(BaseClass):
    num_left_calls = 0
    def call_me(self):
        BaseClass.call_me(self)
        print("Calling method on Left Subclass")
        self.num_left_calls += 1

class RightSubclass(BaseClass):
    num_right_calls = 0
    def call_me(self):
        BaseClass.call_me(self)
        print("Calling method on Right Subclass")
        self.num_right_calls += 1

class Subclass(LeftSubclass, RightSubclass):
    num_sub_calls = 0
    def call_me(self):
        LeftSubclass.call_me(self)
        RightSubclass.call_me(self)
        print("Calling method on Subclass")
        self.num_sub_calls += 1
```

此示例仅确保每个重写的`call_me`方法都直接调用具有相同名称的父方法。 每次将信息打印到屏幕上时，它就会让我们知道方法的调用。 它还会更新类上的静态变量，以显示其已被调用多少次。 如果我们实例化一个`Subclass`对象并对其调用一次，则将得到以下输出：

```py
>>> s = Subclass()
>>> s.call_me()
Calling method on Base Class
Calling method on Left Subclass
Calling method on Base Class
Calling method on Right Subclass
Calling method on Subclass
>>> print(
... s.num_sub_calls,
... s.num_left_calls,
... s.num_right_calls,
... s.num_base_calls)
1 1 1 2

```

因此，我们可以清楚地看到基类的`call_me`方法被调用了两次。 如果该方法进行两次实际工作（例如存入银行帐户），则可能会导致一些隐患。

多重继承要记住的是，我们只想在类层次结构中调用“ next”方法，而不是“ parent”方法。 实际上，该下一个方法可能不在当前类的父级或祖先上。 `super`关键字再次帮助我们。 实际上，`super`最初是为了使复杂形式的多重继承而开发的。 这是使用`super`编写的相同代码：

```py
class BaseClass:
    num_base_calls = 0
    def call_me(self):
        print("Calling method on Base Class")
        self.num_base_calls += 1

class LeftSubclass(BaseClass):
    num_left_calls = 0
    def call_me(self):
        super().call_me()
        print("Calling method on Left Subclass")
        self.num_left_calls += 1

class RightSubclass(BaseClass):
    num_right_calls = 0
    def call_me(self):
        super().call_me()
        print("Calling method on Right Subclass")
        self.num_right_calls += 1

class Subclass(LeftSubclass, RightSubclass):
    num_sub_calls = 0
    def call_me(self):
        super().call_me()
        print("Calling method on Subclass")
        self.num_sub_calls += 1
```

的更改很小。 我们简单地将朴素的直接调用替换为对`super()`的调用，尽管底部子类仅调用`super`一次，而不必对左右两个调用。 更改很简单，但是请看一下执行时的区别：

```py
>>> s = Subclass()
>>> s.call_me()
Calling method on Base Class
Calling method on Right Subclass
Calling method on Left Subclass
Calling method on Subclass
>>> print(s.num_sub_calls, s.num_left_calls, s.num_right_calls,
s.num_base_calls)
1 1 1 1

```

看起来不错，我们的基本方法仅被调用一次。 但是`super()`在这里实际上在做什么？ 由于`print`语句是在`super`调用之后执行的，因此打印输出按实际执行每种方法的顺序排列。 让我们从后到前查看输出，看看谁在打电话。

首先，`Subclass`的`call_me`调用`super().call_me()`，碰巧引用了`LeftSubclass.call_me()`。 然后`LeftSubclass.call_me()`方法调用`super().call_me()`，但是在这种情况下，`super()`指的是`RightSubclass.call_me()`。

请特别注意：`super`调用是*而不是*调用`LeftSubclass`的超类（即`BaseClass`）上的方法。 而是，它正在调用`RightSubclass`，即使它不是`LeftSubclass`的直接父代！ 这是*下一个*方法，而不是父方法。 然后`RightSubclass`调用`BaseClass`，并且`super`调用已确保类层次结构中的每个方法执行一次。

## 不同的参数集

回到`Friend`多继承示例时，使事情变得复杂。 在`Friend`的`__init__`方法中，我们最初为两个父类*调用了带有不同参数集*的`__init__`：

```py
Contact.__init__(self, name, email)
AddressHolder.__init__(self, street, city, state, code)
```

使用`super`时如何管理不同的参数集？ 我们不一定知道哪个类`super`将首先尝试初始化。 即使我们做到了，我们也需要一种传递“额外”参数的方法，以便在其他子类上对`super`的后续调用接收正确的参数。

具体来说，如果第一次调用`super`将`name`和`email`参数传递给`Contact.__init__`，然后`Contact.__init__`然后调用`super`，则它必须能够将与地址相关的参数传递给“ 下一个”方法，即`AddressHolder.__init__`。

每当我们想用相同的名称但使用不同的参数集调用超类方法时，这就是一个问题。 通常，您唯一想调用带有完全不同的参数集的超类的时间是`__init__`，就像我们在这里所做的那样。 即使使用常规方法，我们也可能希望添加仅对一个子类或一组子类有意义的可选参数。

可悲的是，解决此问题的唯一方法是从头开始进行规划。 我们必须设计我们的基类参数列表，以接受每个子类实现不需要的任何参数的关键字参数。 最后，我们必须确保该方法自由地接受意外的参数，并将其传递给其`super`调用，以防它们对于继承顺序中的后续方法是必需的。

Python 的功能参数语法提供了我们执行此操作所需的所有工具，但它使整体代码显得笨重。 看一下`Friend`多重继承代码的正确版本：

```py
class Contact:
    all_contacts = []

    def __init__(self, name='', email='', **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.email = email
        self.all_contacts.append(self)

class AddressHolder:
    def __init__(self, street='', city='', state='', code='',
            **kwargs):
        super().__init__(**kwargs)
        self.street = street
        self.city = city
        self.state = state
        self.code = code

class Friend(Contact, AddressHolder):
    def __init__(self, phone='', **kwargs):
        super().__init__(**kwargs)
        self.phone = phone
```

通过将空字符串作为默认值，我们将所有参数更改为关键字参数。 我们还确保包含一个`**kwargs`参数，以捕获我们的特定方法不知道如何处理的任何其他参数。 它将通过`super`调用将这些参数传递到下一个类。

### 注意

如果您不熟悉`**kwargs`语法，则它基本上会收集传递到该方法中但未在参数列表中明确列出的所有关键字参数。 这些参数存储在名为`kwargs`的字典中（我们可以随意调用该变量，但习惯上建议使用`kw`或`kwargs`）。 当我们使用`**kwargs`语法调用其他方法（例如`super().__init__`）时，它将解压缩字典并将结果作为常规关键字参数传递给该方法。 我们将在第 7 章和 *Python 面向对象的快捷方式*中对此进行详细介绍。

前面的示例完成了应做的工作。 但是它开始显得凌乱，并且已经很难回答以下问题：*我们需要将哪些参数传递给* `Friend.__init__`？ 对于任何打算使用该类的人来说，这是最重要的问题，因此应在方法中添加一个文档字符串以说明正在发生的情况。

此外，如果我们想*重用父类中的*变量，那么即使实现也不足够。 当我们将`**kwargs`变量传递给`super`时，词典不包含任何作为显式关键字参数包含的变量。 例如，在`Friend.__init__`中，对`super`的调用在`kwargs`词典中没有`phone`。 如果其他任何类需要`phone`参数，则需要确保它在传递的字典中。 更糟糕的是，如果我们忘记这样做，那么调试将很困难，因为超类不会抱怨，而只会将默认值（在这种情况下为空字符串）分配给变量。

有几种方法可以确保将变量向上传递。 出于某种原因，假设`Contact`类确实需要使用`phone`参数进行初始化，并且`Friend`类也需要对其进行访问。 我们可以执行以下任一操作：

*   不要将`phone`用作显式关键字参数。 而是将其保留在`kwargs`词典中。 `Friend`可以使用语法`kwargs['phone']`进行查找。 当它将`**kwargs`传递给`super`呼叫时，`phone`仍将在词典中。
*   将`phone`设为显式关键字参数，但使用标准字典语法`kwargs['phone'] = phone`将`kwargs`字典传递给`super`之前，先对其进行更新。
*   使`phone`为显式关键字参数，但使用`kwargs.update`方法更新`kwargs`词典。 如果您有几个要更新的参数，这将很有用。 您可以使用`dict(phone=phone)`构造函数或字典语法`{'phone': phone}`创建传递到`update`的字典。
*   使`phone`为显式关键字参数，但使用语法`super().__init__(phone=phone, **kwargs)`将其显式传递给超级调用。

我们已经讨论了 Python 中涉及多重继承的许多警告。 当我们需要考虑所有可能的情况时，我们必须为它们做计划，我们的代码会变得混乱。 基本的多重继承可能很方便，但是在很多情况下，我们可能希望选择一种更透明的方式来组合两个不同的类，通常使用合成或我们将在第 10 章中介绍的一种设计模式。 ，“Python 设计模式 I” 和第 11 章，“Python 设计模式 II”。

# 多态性

在第 1 章，“面向对象设计”中向我们介绍了多态性。 这是一个很简单的名字，描述了一个简单的概念：根据所使用的子类而发生不同的行为，而不必明确知道该子类实际上是什么。 例如，假设有一个播放音频文件的程序。 媒体播放器可能需要先加载`AudioFile`对象，然后再加载`play`。 我们在对象上放置了`play()`方法，该方法负责解压缩或提取音频并将其路由到声卡和扬声器。 播放`AudioFile`的操作可能很简单，例如：

```py
audio_file.play()
```

但是，对于不同类型的文件，解压缩和提取音频文件的过程非常不同。 `.wav`文件未压缩地存储，而`.mp3`，`.wma`和`.ogg`文件均具有完全不同的压缩算法。

我们可以将继承与多态一起使用以简化设计。 每种类型的文件都可以由`AudioFile`的不同子类表示，例如`WavFile`，`MP3File`。 每个文件都有一个`play()`方法，但是对于每个文件，该方法将以不同的方式实现，以确保遵循正确的提取过程。 媒体播放器对象将永远不需要知道它所指的是`AudioFile`的哪个子类。 它只是调用`play()`并以多态方式让对象处理播放的实际细节。 让我们看一个快速的骨架，显示它的外观：

```py
class AudioFile:
    def __init__(self, filename):
        if not filename.endswith(self.ext):
            raise Exception("Invalid file format")

        self.filename = filename

class MP3File(AudioFile):
    ext = "mp3"
    def play(self):
        print("playing {} as mp3".format(self.filename))

class WavFile(AudioFile):
    ext = "wav"
    def play(self):
        print("playing {} as wav".format(self.filename))

class OggFile(AudioFile):
    ext = "ogg"
    def play(self):
        print("playing {} as ogg".format(self.filename))
```

所有音频文件都会检查以确保在初始化时给出了有效的扩展名。 但是您是否注意到父类中的`__init__`方法如何能够从不同的子类访问`ext`类变量？ 那就是工作中的多态性。 如果文件名后缀名不正确，则会引发异常（下一章将详细讨论例外情况）。 `AudioFile`实际上没有存储对`ext`变量的引用这一事实并不能阻止它能够在子类上对其进行访问。

此外，`AudioFile`的每个子类都以不同的方式实现`play()`（此示例实际上并未播放音乐；音频压缩算法确实值得一个单独的模块！）。 这也是行动中的多态性。 媒体播放器可以使用完全相同的代码来播放文件，而不管文件的类型是什么。 它不在乎它正在查看`AudioFile`的哪个子类。 解压缩音频文件的详细信息封装在[H​​TG4]中。 如果我们测试此示例，它将按我们希望的那样工作：

```py
>>> ogg = OggFile("myfile.ogg")
>>> ogg.play()
playing myfile.ogg as ogg
>>> mp3 = MP3File("myfile.mp3")
>>> mp3.play()
playing myfile.mp3 as mp3
>>> not_an_mp3 = MP3File("myfile.ogg")
Traceback (most recent call last):
 File "<stdin>", line 1, in <module>
 File "polymorphic_audio.py", line 4, in __init__
 raise Exception("Invalid file format")
Exception: Invalid file format

```

看看`AudioFile.__init__`如何能够在不真正知道其引用的子类的情况下检查文件类型？

实际上，多态是有关面向对象编程的最酷的事情之一，它使某些编程设计变得显而易见，而这在早期的范例中是不可能的。 但是，由于使用鸭子输入，Python 使多态性变得不那么酷。 在 Python 中进行鸭子输入可以让我们使用*任何提供所需行为的*对象，而不必强制其成为子类。 Python 的动态特性使其变得无关紧要。 以下示例未扩展`AudioFile`，但可以使用完全相同的接口在 Python 中进行交互：

```py
class FlacFile:
    def __init__(self, filename):
        if not filename.endswith(".flac"):
            raise Exception("Invalid file format")

        self.filename = filename

    def play(self):
        print("playing {} as flac".format(self.filename))
```

我们的媒体播放器可以像扩展`AudioFile`一样轻松地播放此对象。

多态性是在许多面向对象的上下文中使用继承的最重要原因之一。 由于任何提供正确接口的对象都可以在 Python 中互换使用，因此减少了对多态通用超类的需求。 继承对于共享代码仍然有用，但是，如果要共享的只是公共接口，那么只需要鸭子输入即可。 对继承的需求的减少也减少了对多重继承的需求。 通常，当多重继承看起来是一个有效的解决方案时，我们可以使用鸭子类型来模仿多个超类之一。

当然，仅仅因为一个对象满足特定的接口（通过提供所需的方法或属性）并不意味着它会在所有情况下都能正常工作。 它必须以整个系统中有意义的方式实现该接口。 仅仅因为对象提供`play()`方法并不意味着它会自动与媒体播放器一起使用。 例如，我们来自第 1 章，“面向对象设计”的国际象棋 AI 对象可能具有移动棋子的`play()`方法。 即使它满足界面要求，但如果我们尝试将其插入媒体播放器，则此类可能会以惊人的方式中断！

鸭子类型的另一个有用的功能是，鸭子类型的对象仅需要提供那些实际上已被访问的方法和属性。 例如，如果我们需要创建一个假文件对象来读取数据，则可以创建一个具有`read()`方法的新对象； 如果将要与该对象交互的代码仅从文件中读取，则不必重写`write`方法。 更简洁地说，鸭子类型不需要提供可用对象的整个接口，它只需要满足实际访问的接口即可。

# 抽象基类

鸭子类型很有用，但要预先告知某个类是否将满足您所需的协议并不总是那么容易。 因此，Python 引入了抽象基类的思想。 **抽象基类**或 **ABC** 定义了一个类必须实现的一组方法和属性，才能被视为该类的鸭子型实例。 该类可以扩展抽象基类本身，以便用作该类的实例，但是它必须提供所有适当的方法。

在实践中，几乎没有必要创建新的抽象基类，但我们可能会发现实现现有 ABC 实例的机会。 我们将首先介绍如何实现 ABC，然后简要介绍如何在需要时创建自己的 ABC。

## 使用抽象基类

Python 标准库中存在的大多数抽象基类都位于`collections`模块中。 最简单的一种是`Container`类。 让我们在 Python 解释器中检查它，以了解此类需要哪些方法：

```py
>>> from collections import Container
>>> Container.__abstractmethods__
frozenset(['__contains__'])
```

因此，`Container`类只有一种需要实现的抽象方法`__contains__`。 您可以发出`help(Container.__contains__)`来查看功能签名的外观：

```py
Help on method __contains__ in module _abcoll:__contains__(self, x) unbound _abcoll.Container method
```

因此，我们看到`__contains__`需要接受一个参数。 不幸的是，帮助文件没有告诉我们该参数应该是什么，但是从 ABC 的名称和它实现的单个方法中可以很明显地看出，该参数是用户正在检查以查看容器是否容纳的值 。

该方法由`list`，`str`和`dict`实现，以指示给定值是否在该数据结构中。 但是，我们还可以定义一个愚蠢的容器，该容器告诉我们给定值是否在奇数整数集中：

```py
class OddContainer:
    def __contains__(self, x):
        if not isinstance(x, int) or not x % 2:
            return False
        return True
```

现在，我们可以实例化`OddContainer`对象并确定，即使我们没有扩展`Container`，类*也是* `Container`对象：

```py
>>> from collections import Container
>>> odd_container = OddContainer()
>>> isinstance(odd_container, Container)
True
>>> issubclass(OddContainer, Container)
True
```

这就是为什么鸭子类型比经典多态性更出色的原因。 我们可以创建*是*关系，而没有使用继承（或更糟糕的是，多重继承）的开销。

关于`Container` ABC 的有趣之处在于，实现它的任何类都可以免费使用`in`关键字。 实际上，`in`只是委派给`__contains__`方法的语法糖。 具有`__contains__`方法的任何类都是`Container`，因此可以通过`in`关键字查询，例如：

```py
>>> 1 in odd_container
True
>>> 2 in odd_container
False
>>> 3 in odd_container
True
>>> "a string" in odd_container
False
```

## 创建抽象基类

正如我们先前看到的一样，不必具有抽象基类来启用鸭子类型。 但是，想象一下我们正在创建一个带有第三方插件的媒体播放器。 在这种情况下，建议创建一个抽象基类以记录第三方插件应提供的 API。 `abc`模块提供了执行此操作所需的工具，但是我会提前警告您，这需要一些 Python 最神秘的概念：

```py
import abc

class MediaLoader(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def play(self):
        pass

    @abc.abstractproperty
    def ext(self):
        pass

    @classmethod
    def __subclasshook__(cls, C):
        if cls is MediaLoader:
            attrs = set(dir(C))
            if set(cls.__abstractmethods__) <= attrs:
                return True

        return NotImplemented
```

这是一个复杂的示例，其中包含一些 Python 功能，直到本模块稍后再介绍。 为了完整起见，此处包含了它，但是您无需了解所有内容就可以了解如何创建自己的 ABC。

第一个奇怪的是传递给类的`metaclass`关键字参数，通常您会在该类中看到父类的列表。 这是从元类编程的神秘艺术中很少使用的构造。 我们不会在本模块中介绍元类，因此您所需要知道的是，通过分配`ABCMeta`元类，您可以赋予您的类超能力（或至少是超类）能力。

接下来，我们看到`@abc.abstractmethod`和`@abc.abstractproperty`构造。 这些是 Python 装饰器。 我们将在第 5 章，“何时使用面向对象编程”中讨论这些内容。 现在，只知道通过将方法或属性标记为抽象，就可以说明该类的任何子类都必须实现该方法或提供该属性，才能被视为该类的适当成员。

看看如果实现提供或不提供这些属性的子类会发生什么：

```py
>>> class Wav(MediaLoader):
...     pass
...
>>> x = Wav()
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
TypeError: Can't instantiate abstract class Wav with abstract methods ext, play
>>> class Ogg(MediaLoader):
...     ext = '.ogg'
...     def play(self):
...         pass
...
>>> o = Ogg()
```

由于`Wav`类无法实现抽象属性，因此无法实例化该类。 该类仍然是合法的抽象类，但是您必须对其进行子类化才能实际执行任何操作。 `Ogg`类提供这两个属性，因此它可以干净地实例化。

回到`MediaLoader` ABC，我们来剖析`__subclasshook__`方法。 基本上是说，提供该 ABC 所有抽象属性的具体实现的任何类都应被视为`MediaLoader`的子类，即使它实际上不是从`MediaLoader`类继承的也是如此。

更常见的面向对象语言在接口和类的实现之间有着清晰的分隔。 例如，某些语言提供了显式的`interface`关键字，该关键字使我们可以定义类必须具有的方法而无需任何实现。 在这样的环境中，抽象类是一种既提供接口又提供某些但不是全部方法的具体实现的类。 任何类都可以明确声明其实现了给定的接口。

Python 的 ABC 有助于提供接口功能，而不会影响鸭子输入的好处。

## 揭开魔术的神秘面纱

如果要创建满足此特定约定的抽象类，则可以在不了解子类代码的情况下复制并粘贴。 我们将在整个模块中介绍大多数不寻常的语法，但让我们逐行对其进行概述。

```py
    @classmethod
```

该装饰器将该方法标记为类方法。 本质上说，可以在类上而不是实例化对象上调用该方法：

```py
    def __subclasshook__(cls, C):
```

这定义了`__subclasshook__`类方法。 Python 解释器调用此特殊方法来回答问题：*类* `C` *是此类的子类吗？*

```py
        if cls is MediaLoader:
```

我们检查该方法是否专门在此类上调用，而不是说此类的子类。 例如，这可以防止`Wav`类被视为`Ogg`类的父类：

```py
            attrs = set(dir(C))
```

此行所做的只是获取类具有的方法和属性集，包括其类层次结构中的所有父类：

```py
            if set(cls.__abstractmethods__) <= attrs:
```

此行使用集合符号来查看是否在候选类中提供了此类中的抽象方法集。 请注意，它不会检查方法是否已实现，即使它们在那里也是如此。 因此，一个类有可能是一个子类，而本身仍然是一个抽象类。

```py
                return True
```

如果提供了所有抽象方法，则候选类为该类的子类，我们返回`True`。 该方法可以合法地返回三个值之一：`True`，`False`或`NotImplemented`。 `True`和`False`表示该类绝对是该类的子类：

```py
        return NotImplemented
```

如果未满足任何条件（即，该类不是`MediaLoader`或未提供所有抽象方法），则返回`NotImplemented`。 这告诉 Python 机制使用默认机制（候选类是否显式扩展了该类？）进行子类检测。

简而言之，我们现在可以将`Ogg`类定义为`MediaLoader`类的子类，而无需实际扩展`MediaLoader`类：

```py
>>> class Ogg():
...     ext = '.ogg'
...     def play(self):
...         print("this will play an ogg file")
...
>>> issubclass(Ogg, MediaLoader)
True
>>> isinstance(Ogg(), MediaLoader)
True
```

# 案例研究

让我们尝试将我们学到的所有内容与一个更大的例子联系在一起。 我们将设计一个简单的房地产应用，使代理可以管理可用于购买或出租的财产。 物业将分为两种：公寓和房屋。 代理需要能够输入有关新属性的一些相关详细信息，列出所有当前可用的属性，以及将属性标记为已出售或已租赁。 为简便起见，我们无需担心在出售后编辑属性详细信息或重新激活属性的问题。

该项目将允许代理使用 Python 解释器提示与对象进行交互。 在这个图形用户界面和 Web 应用的世界里，您可能想知道为什么我们要创建这种老式外观的程序。 简而言之，窗口化程序和 Web 应用都需要大量开销知识和样板代码才能使它们执行所需的操作。 如果我们使用这两种范例中的任何一种来开发软件，那么我们会在 GUI 编程或 Web 编程中迷失方向，以至于忽视了我们试图掌握的面向对象原理。

幸运的是，大多数 GUI 和 Web 框架都采用了面向对象的方法，而我们现在正在研究的原理将有助于将来理解这些系统。 我们将在第 13 章，“并发”中简要讨论它们，但是完整的细节远远超出了单个模块的范围。

查看我们的要求，似乎有很多名词可以表示系统中的对象类别。 显然，我们需要表示一个属性。 房屋和公寓可能需要单独的课程。 租金和购买似乎也需要单独的代表。 由于我们现在专注于继承，因此我们将研究使用继承或多继承共享行为的方法。

`House`和`Apartment`都是属性的类型，因此`Property`可以是这两个类的超类。 `Rental`和`Purchase`需要额外考虑； 如果我们使用继承，则需要有单独的类，例如`HouseRental`和`HousePurchase`，并使用多重继承将它们组合在一起。 与基于合成或基于关联的设计相比，这感觉有些笨拙，但是让我们一起运行它，看看我们提出了什么。

现在，哪些属性可能与`Property`类相关联？ 无论是公寓还是房屋，大多数人都想知道平方英尺，卧室数量和浴室数量。 （还有许多其他可以建模的属性，但对于我们的原型，我们将使其保持简单。）

如果该物业是一栋房屋，它将要宣传楼层的数量，是否有车库（有，独立或无人车库）以及院子是否被围起来。 公寓将要指出是否有阳台，以及洗衣房是套房式，硬币式还是非现场式。

这两种属性类型都需要一种方法来显示该属性的特征。 目前，没有其他行为可见。

租赁物业将需要存储每月的租金，物业是否配备，是否包括公用事业，如果不包含，则应估算其租金。 购买的物业将需要存储购买价格和估计的年度物业税。 对于我们的应用，我们只需要显示此数据，因此我们只需添加类似于其他类中使用的`display()`方法就可以摆脱困境。

最后，我们需要一个`Agent`对象，该对象保存所有属性的列表，显示这些属性，并允许我们创建新属性。 创建属性将需要提示用户输入每种属性类型的相关详细信息。 可以在`Agent`对象中完成此操作，但随后`Agent`将需要了解许多有关属性类型的信息。 这没有利用多态性。 另一种选择是将提示放在每个类的初始化器甚至是构造函数中，但这将不允许将来在 GUI 或 Web 应用中应用这些类。 更好的主意是创建一个执行提示并返回提示参数字典的静态方法。 然后，`Agent`所要做的就是提示用户输入财产的类型和付款方式，并要求正确的类实例化自己。

大量的设计！ 下面的类图可以更清晰地传达我们的设计决策：

![Case study](img/8781OS_3_03.jpg)

哇，那是很多继承箭头！ 我认为不加箭头就不可能添加另一个继承级别。 即使在设计阶段，多重继承也很麻烦。

这些类最棘手的方面将是确保在继承层次结构中调用超类方法。 让我们从`Property`实现开始：

```py
class Property:
    def __init__(self, square_feet='', beds='',
            baths='', **kwargs):
        super().__init__(**kwargs)
        self.square_feet = square_feet
        self.num_bedrooms = beds
        self.num_baths = baths

    def display(self):
        print("PROPERTY DETAILS")
        print("================")
        print("square footage: {}".format(self.square_feet))
        print("bedrooms: {}".format(self.num_bedrooms))
        print("bathrooms: {}".format(self.num_baths))
        print()

    def prompt_init():
        return dict(square_feet=input("Enter the square feet: "),
                beds=input("Enter number of bedrooms: "),
                baths=input("Enter number of baths: "))
    prompt_init = staticmethod(prompt_init)

```

这个类非常简单。 我们已经在`__init__`中添加了额外的`**kwargs`参数，因为我们知道它将在多重继承情况下使用。 如果我们不是多重继承链中的最后一个调用，我们还包括了对`super().__init__`的调用。 在这种情况下，我们正在*使用*关键字参数，因为我们知道在继承层次结构的其他级别上将不需要它们。

我们在`prompt_init`方法中看到了一些新内容。 最初创建此方法后，立即将其变为静态方法。 静态方法仅与类（类似于类变量）相关联，而不与特定的对象实例相关联。 因此，它们没有`self`参数。 因此，`super`关键字将不起作用（没有父对象，只有父类），因此我们直接在父类上直接调用 static 方法。 此方法使用 Python `dict`构造函数创建一个值字典，该值可以传递到`__init__`中。 调用`input`会提示每个键的值。

`Apartment`类扩展了`Property`，其结构类似：

```py
class Apartment(Property):
    valid_laundries = ("coin", "ensuite", "none")
    valid_balconies = ("yes", "no", "solarium")

    def __init__(self, balcony='', laundry='', **kwargs):
        super().__init__(**kwargs)
        self.balcony = balcony
        self.laundry = laundry

    def display(self):
        super().display()
        print("APARTMENT DETAILS")
        print("laundry: %s" % self.laundry)
        print("has balcony: %s" % self.balcony)

    def prompt_init():
        parent_init = Property.prompt_init()
        laundry = ''
        while laundry.lower() not in \
                Apartment.valid_laundries:
            laundry = input("What laundry facilities does "
                    "the property have? ({})".format(
                    ", ".join(Apartment.valid_laundries)))
        balcony = ''
        while balcony.lower() not in \
                Apartment.valid_balconies:
            balcony = input(
                "Does the property have a balcony? "
                "({})".format(
                ", ".join(Apartment.valid_balconies)))
        parent_init.update({
            "laundry": laundry,
            "balcony": balcony
        })
        return parent_init
    prompt_init = staticmethod(prompt_init)
```

`display()`和`__init__()`方法使用`super()`调用各自的父类方法，以确保`Property`类被正确初始化。

`prompt_init`静态方法现在从父类获取字典值，然后添加其自身的一些其他值。 它调用`dict.update`方法将新的字典值合并到第一个字典中。 但是，该`prompt_init`方法看起来很丑陋。 它会循环两次，直到用户使用结构相似的代码但变量不同来输入有效输入。 提取此验证逻辑会很好，因此我们只能将其维护在一个位置。 这对以后的课程也很有用。

关于继承的所有讨论，我们可能会认为这是使用 mixin 的好地方。 相反，我们有机会研究继承不是最佳解决方案的情况。 我们要创建的方法将在静态方法中使用。 如果要从提供验证功能的类继承，则也必须以不访问该类上任何实例变量的静态方法提供该功能。 如果它不访问任何实例变量，那么使其完全成为类的意义何在？ 为什么我们不只是将此验证功能设为接受输入字符串和有效答案列表的模块级功能，而只保留它呢？

让我们探讨一下此验证函数的外观：

```py
def get_valid_input(input_string, valid_options):
    input_string += " ({}) ".format(", ".join(valid_options))
    response = input(input_string)
    while response.lower() not in valid_options:
        response = input(input_string)
    return response
```

我们可以在解释器中测试此功能，而与我们一直在研究的所有其他类无关。 这是一个好兆头，这意味着我们设计的不同部分之间没有紧密耦合，并且可以在不影响其他代码部分的情况下独立进行改进。

```py
>>> get_valid_input("what laundry?", ("coin", "ensuite", "none"))
what laundry? (coin, ensuite, none) hi
what laundry? (coin, ensuite, none) COIN
'COIN'

```

现在，让我们快速更新`Apartment.prompt_init`方法以使用此新功能进行验证：

```py
    def prompt_init():
        parent_init = Property.prompt_init()
        laundry = get_valid_input(
                "What laundry facilities does "
                "the property have? ",
                Apartment.valid_laundries)
        balcony = get_valid_input(
            "Does the property have a balcony? ",
            Apartment.valid_balconies)
        parent_init.update({
            "laundry": laundry,
            "balcony": balcony
        })
        return parent_init
    prompt_init = staticmethod(prompt_init)
```

与我们的原始版本相比，更易于阅读（和维护！）。 现在我们准备构建`House`类。 此类具有与`Apartment`并行的结构，但是引用了不同的提示和变量：

```py
class House(Property):
    valid_garage = ("attached", "detached", "none")
    valid_fenced = ("yes", "no")

    def __init__(self, num_stories='',
            garage='', fenced='', **kwargs):
        super().__init__(**kwargs)
        self.garage = garage
        self.fenced = fenced
        self.num_stories = num_stories

    def display(self):
        super().display()
        print("HOUSE DETAILS")
        print("# of stories: {}".format(self.num_stories))
        print("garage: {}".format(self.garage))
        print("fenced yard: {}".format(self.fenced))

    def prompt_init():
        parent_init = Property.prompt_init()
        fenced = get_valid_input("Is the yard fenced? ",
                    House.valid_fenced)
        garage = get_valid_input("Is there a garage? ",
                House.valid_garage)
        num_stories = input("How many stories? ")

        parent_init.update({
            "fenced": fenced,
            "garage": garage,
            "num_stories": num_stories
        })
        return parent_init
    prompt_init = staticmethod(prompt_init)
```

这里没有的新知识，因此让我们继续进行`Purchase`和`Rental`类。 尽管目的明显不同，但它们在设计上也与我们刚刚讨论的目的相似：

```py
class Purchase:
    def __init__(self, price='', taxes='', **kwargs):
        super().__init__(**kwargs)
        self.price = price
        self.taxes = taxes

    def display(self):
        super().display()
        print("PURCHASE DETAILS")
        print("selling price: {}".format(self.price))
        print("estimated taxes: {}".format(self.taxes))

    def prompt_init():
        return dict(
            price=input("What is the selling price? "),
            taxes=input("What are the estimated taxes? "))
    prompt_init = staticmethod(prompt_init)

class Rental:
    def __init__(self, furnished='', utilities='',
            rent='', **kwargs):
        super().__init__(**kwargs)
        self.furnished = furnished
        self.rent = rent
        self.utilities = utilities

    def display(self):
        super().display()
        print("RENTAL DETAILS")
        print("rent: {}".format(self.rent))
        print("estimated utilities: {}".format(
            self.utilities))
        print("furnished: {}".format(self.furnished))

    def prompt_init():
        return dict(
            rent=input("What is the monthly rent? "),
            utilities=input(
                "What are the estimated utilities? "),
            furnished = get_valid_input(
                "Is the property furnished? ",
                    ("yes", "no")))
    prompt_init = staticmethod(prompt_init)
```

这两个类没有超类（`object`除外），但我们仍将其称为`super().__init__`，因为它们将与其他类结合使用，并且我们不知道`super`的顺序 将进行调用。该接口类似于`House`和`Apartment`所使用的接口，当我们将这四个类的功能组合到单独的子类中时，该接口非常有用。 例如：

```py
class HouseRental(Rental, House):
    def prompt_init():
        init = House.prompt_init()
        init.update(Rental.prompt_init())
        return init
    prompt_init = staticmethod(prompt_init)
```

这有点令人惊讶，因为类本身既没有`__init__`也没有`display`方法！ 因为两个父类都在这些方法中适当地调用了`super`，所以我们仅需扩展这些类，并且这些类将以正确的顺序运行。 当然，`prompt_init`并非如此，因为它是一个静态方法，不会调用`super`，因此我们明确实现了这一方法。 在编写其他三个组合之前，我们应该测试此类以确保其行为正确：

```py
>>> init = HouseRental.prompt_init()
Enter the square feet: 1
Enter number of bedrooms: 2
Enter number of baths: 3
Is the yard fenced?  (yes, no) no
Is there a garage?  (attached, detached, none) none
How many stories? 4
What is the monthly rent? 5
What are the estimated utilities? 6
Is the property furnished?  (yes, no) no
>>> house = HouseRental(**init)
>>> house.display()
PROPERTY DETAILS
================
square footage: 1
bedrooms: 2
bathrooms: 3

HOUSE DETAILS
# of stories: 4
garage: none
fenced yard: no

RENTAL DETAILS
rent: 5
estimated utilities: 6
furnished: no

```

看起来工作正常。 `prompt_init`方法正在提示所有超类的初始化程序，`display()`也正在协作调用所有三个超类。

### 注意

前面示例中继承类的顺序很重要。 如果我们写的是`class HouseRental(House, Rental)`而不是`class HouseRental(Rental, House)`，则`display()`不会调用`Rental.display()`！ 在我们的`HouseRental`版本上调用`display`时，它指的是该方法的`Rental`版本，该方法调用`super.display()`获得`House`版本，然后再次调用`super.display()`获得属性版本。 。 如果我们将其反转，则`display`将引用`House`类的`display()`。 调用 super 时，它将调用`Property`父类的方法。 但是`Property`在其`display`方法中没有对`super`的调用。 这意味着将不会调用`Rental`类的`display`方法！ 通过按照我们执行的顺序放置继承列表，我们确保`Rental`调用`super`，这将处理层次结构的`House`端。 您可能以为我们可以在`Property.display()`中添加`super`调用，但这会失败，因为`Property`的下一个超类是`object`，并且`object`没有`display`方法。 解决此问题的另一种方法是允许`Rental`和`Purchase`扩展`Property`类，而不是直接从`object`派生。 （或者我们可以动态修改方法的解析顺序，但这超出了本模块的范围。）

现在我们已经测试了它，我们准备创建其余的组合子类：

```py
class ApartmentRental(Rental, Apartment):
    def prompt_init():
        init = Apartment.prompt_init()
        init.update(Rental.prompt_init())
        return init
    prompt_init = staticmethod(prompt_init)

class ApartmentPurchase(Purchase, Apartment):
    def prompt_init():
        init = Apartment.prompt_init()
        init.update(Purchase.prompt_init())
        return init
    prompt_init = staticmethod(prompt_init)

class HousePurchase(Purchase, House):
    def prompt_init():
        init = House.prompt_init()
        init.update(Purchase.prompt_init())
        return init
    prompt_init = staticmethod(prompt_init)
```

那应该是我们最紧张的设计！ 现在，我们要做的就是创建`Agent`类，该类负责创建新列表并显示现有列表。 让我们从更简单的属性存储和列出开始：

```py
class Agent:
    def __init__(self):
        self.property_list = []

    def display_properties(self):
        for property in self.property_list:
            property.display()
```

添加属性需要首先查询属性的类型以及属性是用于购买还是出租。 我们可以通过显示一个简单的菜单来做到这一点。 一旦确定了这一点，我们就可以使用我们已经开发的`prompt_init`层次结构提取正确的子类并提示所有详细信息。 听起来很简单？ 它是。 首先，向`Agent`类添加一个字典类变量：

```py
    type_map = {
        ("house", "rental"): HouseRental,
        ("house", "purchase"): HousePurchase,
        ("apartment", "rental"): ApartmentRental,
        ("apartment", "purchase"): ApartmentPurchase
        }
```

那是一些漂亮的有趣的代码。 这是一本字典，其中的键是两个不同字符串的元组，而值是类对象。 类对象？ 是的，可以像*普通*对象或原始数据类型一样，将类传递，重命名并存储在容器中。 使用这个简单的字典，我们可以简单地劫持我们先前的`get_valid_input`方法，以确保获得正确的字典键并查找适当的类，如下所示：

```py
    def add_property(self):
        property_type = get_valid_input(
                "What type of property? ",
                ("house", "apartment")).lower()
        payment_type = get_valid_input(
                "What payment type? ",
                ("purchase", "rental")).lower()

        PropertyClass = self.type_map[
            (property_type, payment_type)]
        init_args = PropertyClass.prompt_init()
        self.property_list.append(PropertyClass(**init_args))

```

这也可能看起来很有趣！ 我们在字典中查找该类，并将其存储在名为`PropertyClass`的变量中。 我们不确切知道哪个类可用，但是该类自己知道，因此我们可以多态调用`prompt_init`以获取适合传递给构造函数的值的字典。 然后，我们使用关键字参数语法将字典转换为参数，并构造新对象以加载正确的数据。

现在，我们的用户可以使用此`Agent`类添加和查看属性列表。 添加功能以将属性标记为可用或不可用，或者编辑和删除属性都不需要太多的工作。 我们的原型现在处于足够好的状态，可以带入房地产`agent`并演示其功能。 这是演示会话的工作方式：

```py
>>> agent = Agent()
>>> agent.add_property()
What type of property?  (house, apartment) house
What payment type?  (purchase, rental) rental
Enter the square feet: 900
Enter number of bedrooms: 2
Enter number of baths: one and a half
Is the yard fenced?  (yes, no) yes
Is there a garage?  (attached, detached, none) detached
How many stories? 1
What is the monthly rent? 1200
What are the estimated utilities? included
Is the property furnished?  (yes, no) no
>>> agent.add_property()
What type of property?  (house, apartment) apartment
What payment type?  (purchase, rental) purchase
Enter the square feet: 800
Enter number of bedrooms: 3
Enter number of baths: 2
What laundry facilities does the property have?  (coin, ensuite,
one) ensuite
Does the property have a balcony? (yes, no, solarium) yes
What is the selling price? $200,000
What are the estimated taxes? 1500
>>> agent.display_properties()
PROPERTY DETAILS
================
square footage: 900
bedrooms: 2
bathrooms: one and a half

HOUSE DETAILS
# of stories: 1
garage: detached
fenced yard: yes
RENTAL DETAILS
rent: 1200
estimated utilities: included
furnished: no
PROPERTY DETAILS
================
square footage: 800
bedrooms: 3
bathrooms: 2

APARTMENT DETAILS
laundry: ensuite
has balcony: yes
PURCHASE DETAILS
selling price: $200,000
estimated taxes: 1500

```

![Case study](img/C_01_04a.jpg)

![Case study](img/C_01_04b.jpg)

![Case study](img/Summary_01_04.jpg)

![Case study](img/Yourprogress_01.jpg)