# 十、迭代器模式

我们已经讨论了多少 Python 的内置和习惯用法，乍一看似乎是非面向对象的，实际上是提供了对主要对象的访问。 在本章中，我们将讨论看起来如此结构化的`for`循环实际上是围绕一组面向对象原理的轻量级包装。 我们还将看到对该语法的各种扩展，这些扩展会自动创建更多类型的对象。 我们将介绍：

*   什么是设计模式
*   迭代器协议-最强大的设计模式之一
*   列表，集合和字典理解
*   发电机和协程

# 设计模式简介

当工程师和建筑师决定建造桥梁，塔楼或建筑物时，他们遵循某些原则以确保结构完整性。 桥梁有多种可能的设计（例如悬架或悬臂），但是如果工程师不使用一种标准设计，并且没有出色的新设计，则他/她设计的桥梁可能会 坍塌。

设计模式旨在将用于正确设计的结构的相同正式定义引入软件工程。 有许多不同的设计模式可以解决不同的一般问题。 创建设计模式的人首先会确定开发人员在各种情况下面临的常见问题。 然后，他们就面向对象设计提出了可能被认为是该问题的理想解决方案的建议。

但是，知道设计模式并选择在我们的软件中使用它并不能保证我们正在创建“正确的”解决方案。 1907 年，魁北克大桥（至今是世界上最长的悬臂桥）在施工完成之前倒塌，因为设计桥梁的工程师严重低估了用于建造它的钢材的重量。 同样，在软件开发中，我们可能会错误地选择或应用设计模式，并创建在正常操作情况下或在超出其原始设计限制的情况下会“崩溃”的软件。

任何一个设计模式都提出了一组以特定方式交互的对象，以解决一般问题。 程序员的职责是识别他们何时面对该问题的特定版本，并在其解决方案中适应常规设计。

在本章中，我们将介绍迭代器设计模式。 这种模式是如此强大和普遍，以至于 Python 开发人员提供了多种语法来访问该模式基础的面向对象原理。 在接下来的两章中，我们将介绍其他设计模式。 其中有些支持语言，有些则没有，但是它们都没有像迭代器模式那样固有地在 Python 编码器的日常生活中发挥作用。

# 迭代器

在典型的模式设计中，迭代器是使用`next()`方法和`done()`方法的对象； 如果序列中没有剩余项目，则后者返回`True`。 在没有内置支持迭代器的编程语言中，迭代器将像这样循环遍历：

```py
while not iterator.done():
    item = iterator.next()
    # do something with the item
```

在 Python 中，迭代是一个特殊功能，因此该方法有一个特殊名称`__next__`。 可以使用内置的`next(iterator)`访问此方法。 迭代器协议使用`StopIteration`而不是`done`方法来通知循环它已完成。 最后，我们具有更具可读性的`for item in iterator`语法，可以实际访问迭代器中的项，而不必搞乱`while`循环。 让我们更详细地看看这些。

## 迭代器协议

`collections.abc`模块中的抽象基类`Iterator`定义了 Python 中的迭代器协议。 如前所述，它必须具有`__next__`方法，`for`循环（以及支持迭代的其他功能）可以调用该方法以从序列中获取新元素。 此外，每个迭代器还必须满足`Iterable`接口。 任何提供`__iter__`方法的类都是可迭代的。 该方法必须返回一个`Iterator`实例，该实例将覆盖该类中的所有元素。 由于迭代器已经在元素上循环，因此其`__iter__`函数通常会返回自身。

这听起来可能有些混乱，所以请看以下示例，但是请注意，这是解决此问题的非常冗长的方法。 它清楚地解释了迭代以及所涉及的两个协议，但是在本章后面，我们将研究几种更易读的方法来获得这种效果：

```py
class CapitalIterable:
    def __init__(self, string):
        self.string = string

    def __iter__(self):
        return CapitalIterator(self.string)

class CapitalIterator:
    def __init__(self, string):
        self.words = [w.capitalize() for w in string.split()]
        self.index = 0

    def __next__(self):
        if self.index == len(self.words):
            raise StopIteration()

        word = self.words[self.index]
        self.index += 1
        return word

    def __iter__(self):
        return self
```

本示例定义了一个`CapitalIterable`类，其工作是循环遍历字符串中的每个单词，并以首字母大写输出它们。 可迭代的大部分工作都传递给`CapitalIterator`实现。 与该迭代器进行交互的规范方法如下：

```py
>>> iterable = CapitalIterable('the quick brown fox jumps over the lazy dog')
>>> iterator = iter(iterable)
>>> while True:
...     try:
...         print(next(iterator))
...     except StopIteration:
...         break
... 
The
Quick
Brown
Fox
Jumps
Over
The
Lazy
Dog

```

此示例首先构造一个可迭代的对象，并从中检索一个迭代器。 区别可能需要解释； 可迭代的对象是具有可以循环的元素的对象。 通常，这些元素可以多次循环，甚至可以同时循环或以重叠代码循环。 另一方面，迭代器表示该迭代器中的特定位置。 有些物品已经消耗掉了，有些则没有。 两个不同的迭代器可能在单词列表中的不同位置，但是任何一个迭代器都只能标记一个位置。

每次在迭代器上调用`next()`时，它都会依次从迭代器中返回另一个令牌。 最终，迭代器将被耗尽（不再有其他要返回的元素），在这种情况下，将引发`Stopiteration`，我们就跳出了循环。

当然，我们已经知道从可迭代对象构造迭代器的语法要简单得多：

```py
>>> for i in iterable:
...     print(i)
... 
The
Quick
Brown
Fox
Jumps
Over
The
Lazy
Dog

```

正如您所看到的，尽管`for`语句看上去并不十分面向对象，但实际上是某些明显的面向对象设计原则的捷径。 当我们讨论理解时，请记住这一点，因为它们也似乎与面向对象工具相反。 但是，它们使用与`for`循环完全相同的迭代协议，并且只是另一种捷径。

# 理解

理解是简单但功能强大的语法，使我们能够以最少的一行代码来转换或过滤可迭代的对象。 结果对象可以是完全正常的列表，集合或字典，也可以是可以一次性有效使用的生成器表达式。

## 列表理解

列表推导是 Python 中最强大的工具之一，因此人们倾向于将其视为高级工具。 他们不是。 确实，我已经自由地带着理解力乱扔先前的示例，并假设您会理解它们。 确实，高级程序员经常使用理解能力，但这并不是因为他们很先进，而是因为他们很琐碎，并且能够处理软件开发中最常见的一些操作。

让我们看一下其中的一种常见操作； 即，将项目列表转换为相关项目列表。 具体来说，假设我们只是从文件中读取字符串列表，现在我们想将其转换为整数列表。 我们知道列表中的每个项目都是整数，并且我们希望对这些数字进行一些活动（例如，计算平均值）。 这是一种简单的处理方法：

```py
input_strings = ['1', '5', '28', '131', '3']

output_integers = []
for num in input_strings:
    output_integers.append(int(num))
```

可以正常工作，并且只有三行代码。 如果您不习惯于理解，您甚至可能不会认为它看起来很丑！ 现在，使用列表理解来查看相同的代码：

```py
input_strings = ['1', '5', '28', '131', '3']output_integers = [int(num) for num in input_strings]
```

我们将简化为一行，并且为提高性能，重要的是，我们为列表中的每个项目都删除了`append`方法调用。 总体而言，即使您不习惯理解语法，也很容易知道发生了什么。

与以往一样，方括号表示我们正在创建列表。 该列表内部是一个`for`循环，该循环遍历输入序列中的每个项目。 唯一令人困惑的是列表的开头大括号与`for`循环的开始之间发生的情况。 输入列表中的每项都将在此处发生的任何情况应用于*。 有问题的项目由循环中的`num`变量引用。 因此，它将每个单独的元素转换为`int`数据类型。*

这就是基本列表理解的全部内容。 毕竟它们并不那么先进。 理解是高度优化的 C 代码； 遍历大量项目时，列表理解比`for`循环快得多。 如果仅凭可读性并不是说服人们尽可能多地使用它们的理由，那么应该提高速度。

将列表项转换为一个相关列表并不是我们对列表理解的唯一操作。 我们还可以通过在理解内添加`if`语句来选择排除某些值。 看一看：

```py
output_ints = [int(n) for n in input_strings if len(n) < 3]
```

我将变量的名称从`num`缩短为`n`，并将结果变量缩短为`output_ints`，因此它仍然可以放在一行上。 除此之外，此示例与上一个示例的不同之处是`if len(n) < 3`部分。 此额外的代码排除了两个以上字符的字符串。 `if`语句在`int`函数之前应用，因此它正在测试字符串的长度。 由于我们的输入字符串本质上都是整数，因此它排除了 99 以上的任何数字。现在，所有列出的内容都将包含在内！ 我们使用它们将输入值映射为输出值，并沿途应用过滤器以包含或排除任何满足特定条件的值。

任何可迭代的都可以作为列表理解的输入； 我们可以包装在`for`循环中的所有内容也可以放在理解内。 例如，文本文件是可迭代的。 文件的迭代器上对`__next__`的每次调用都将返回文件的一行。 我们可以使用`zip`函数将制表符分隔的文件（其中第一行是标题行）加载到字典中：

```py
import sys
filename = sys.argv[1]

with open(filename) as file:
    header = file.readline().strip().split('\t')
 contacts = [
 dict(
 zip(header, line.strip().split('\t'))
 ) for line in file
 ]

for contact in contacts:
    print("email: {email} -- {last}, {first}".format(
        **contact))
```

这次，我添加了一些空格以使其更具可读性（列表理解*的*不能放在一行上）。 本示例从压缩的标题创建字典列表，并为文件中的每一行分割行。

嗯什么 如果该代码或解释没有意义，请不要担心。 这有点令人困惑。 列表理解在这里需要做大量工作，并且代码很难理解，阅读和维护。 这个例子表明列表理解并不总是最好的解决方案。 大多数程序员都同意`for`循环比该版本更具可读性。

### 提示

切记：请勿滥用我们提供的工具！ 始终选择适合工作的正确工具，这总是编写可维护的代码。

## 集合和字典理解

理解力不仅限于列表。 我们可以使用带有大括号的类似语法来创建集合，并像一样创建字典。 让我们从集合开始。 创建集合的一种方法是将列表理解包装在`set()`构造函数中，该构造函数将其转换为集合。 但是，当我们可以直接创建一个集合时，为什么要在中间列表上浪费内存呢？

这是一个示例，该示例使用命名元组对作者/标题/类型三合会进行建模，然后检索以特定类型撰写的所有作者的集合：

```py
from collections import namedtuple

Book = namedtuple("Book", "author title genre")
books = [
        Book("Pratchett", "Nightwatch", "fantasy"),
        Book("Pratchett", "Thief Of Time", "fantasy"),
        Book("Le Guin", "The Dispossessed", "scifi"),
        Book("Le Guin", "A Wizard Of Earthsea", "fantasy"),
        Book("Turner", "The Thief", "fantasy"),
        Book("Phillips", "Preston Diamond", "western"),
        Book("Phillips", "Twice Upon A Time", "scifi"),
        ]

fantasy_authors = {
 b.author for b in books if b.genre == 'fantasy'}

```

与演示数据设置相比，突出显示的设置理解力肯定很短！ 如果我们使用列表推导，那么 Terry Pratchett 当然会被列出两次。按原样，集合的性质将删除重复项，最后得到：

```py
>>> fantasy_authors
{'Turner', 'Pratchett', 'Le Guin'}

```

我们可以引入冒号来创建字典理解。 这将使用*键将值转换为字典：*对值。 例如，如果我们知道标题，则快速查找字典中的作者或体裁可能会很有用。 我们可以使用字典理解将标题映射到模块对象：

```py
fantasy_titles = {
        b.title: b for b in books if b.genre == 'fantasy'}
```

现在，我们有了字典，可以使用常规语法按标题查找书籍。

总之，理解不是高级 Python，也不是应避免的“非面向对象”工具。 它们只是从现有序列创建列表，集合或字典的一种更为简洁和优化的语法。

## 生成器表达式

有时我们想处理一个新序列而不将新列表，集合或字典放入系统内存中。 如果我们一次只遍历一个项目，而实际上并不在乎是否要创建最终的容器对象，那么创建该容器就是浪费内存。 一次处理一项时，我们只需要随时将当前对象存储在内存中。 但是，当我们创建一个容器时，所有对象都必须先存储在该容器中，然后才能开始处理它们。

例如，考虑一个处理日志文件的程序。 一个非常简单的日志可能包含以下格式的信息：

```py
Jan 26, 2015 11:25:25    DEBUG        This is a debugging message.
Jan 26, 2015 11:25:36    INFO         This is an information method.
Jan 26, 2015 11:25:46    WARNING      This is a warning. It could be serious.
Jan 26, 2015 11:25:52    WARNING      Another warning sent.
Jan 26, 2015 11:25:59    INFO         Here's some information.
Jan 26, 2015 11:26:13    DEBUG        Debug messages are only useful if you want to figure something out.
Jan 26, 2015 11:26:32    INFO         Information is usually harmless, but helpful.
Jan 26, 2015 11:26:40    WARNING      Warnings should be heeded.
Jan 26, 2015 11:26:54    WARNING      Watch for warnings.
```

流行的 Web 服务器，数据库或电子邮件服务器的日志文件可能包含许多 GB 的数据（我最近不得不从行为异常的系统中清除近 2 TB 的日志）。 如果要处理日志中的每一行，则不能使用列表推导。 它将创建一个包含文件中每一行的列表。 这可能不适合 RAM，并且可能会使计算机屈服，具体取决于操作系统。

如果在日志文件上使用`for`循环，则可以一次处理一行，然后再将下一行读入内存。 如果我们可以使用理解语法来达到相同的效果，那会不好吗？

这就是生成器表达式的用处。它们使用与理解相同的语法，但不会创建最终的容器对象。 要创建生成器表达式，请将理解内容包装在`()`中，而不是`[]`或`{}`中。

以下代码以以前介绍的格式解析日志文件，并输出仅包含`WARNING`行的新日志文件：

```py
import sys

inname = sys.argv[1]
outname = sys.argv[2]

with open(inname) as infile:
    with open(outname, "w") as outfile:
 warnings = (l for l in infile if 'WARNING' in l)
        for l in warnings:
            outfile.write(l)
```

此程序在命令行上使用两个文件名，使用生成器表达式过滤掉警告（在这种情况下，它使用`if`语法，并且不修改该行），然后将警告输出到另一个 文件。 如果我们在示例文件中运行它，则输出如下所示：

```py
Jan 26, 2015 11:25:46    WARNING     This is a warning. It could be serious.
Jan 26, 2015 11:25:52    WARNING     Another warning sent.
Jan 26, 2015 11:26:40    WARNING     Warnings should be heeded.
Jan 26, 2015 11:26:54    WARNING     Watch for warnings.
```

当然，使用这么短的输入文件，我们可以安全地使用列表推导，但是如果文件长数百万行，则生成器表达式将对内存和速度产生巨大影响。

生成器表达式通常在函数调用中最有用。 例如，我们可以在生成器表达式而不是列表上调用`sum`，`min`或`max`，因为这些函数一次处理一个对象。 我们只对结果感兴趣，而不对任何中间容器感兴趣。

通常，应尽可能使用生成器表达式。 如果我们实际上不需要列表，集合或字典，而只需要过滤或转换序列中的项目，则生成器表达式将是最有效的。 如果我们需要知道列表的长度，对结果进行排序，删除重复项或创建字典，则必须使用理解语法。

# 发电机

生成器表达式实际上也是一种理解。 他们将更高级的（这次确实更高级！）生成器语法压缩为一行。 更大的生成器语法看起来甚至比我们所见的面向对象要少，但是我们会再次发现，创建一种对象是一种简单的语法捷径。

让我们进一步介绍日志文件示例。 如果我们要从输出文件中删除`WARNING`列（由于它是多余的：此文件仅包含警告），我们有几种选项，具有不同的可读性。 我们可以使用生成器表达式来做到这一点：

```py
import sys
inname, outname = sys.argv[1:3]

with open(inname) as infile:
    with open(outname, "w") as outfile:
 warnings = (l.replace('\tWARNING', '')
 for l in infile if 'WARNING' in l)
        for l in warnings:
            outfile.write(l)
```

这是完全可读的，尽管我不想使表达式复杂得多。 我们也可以使用普通的`for`循环来完成此操作：

```py
import sys
inname, outname = sys.argv[1:3]

with open(inname) as infile:
    with open(outname, "w") as outfile:
 for l in infile:
 if 'WARNING' in l:
 outfile.write(l.replace('\tWARNING', ''))

```

这是可以维持的，但是在很少的几行中缩进那么多的等级有点难看。 更令人震惊的是，如果我们想对这些行做一些不同的事情，而不仅仅是打印出来，我们还必须复制循环代码和条件代码。 现在让我们考虑一个没有任何捷径的真正面向对象的解决方案：

```py
import sys
inname, outname = sys.argv[1:3]

class WarningFilter:
 def __init__(self, insequence):
 self.insequence = insequence
 def __iter__(self):
 return self
 def __next__(self):
 l = self.insequence.readline()
 while l and 'WARNING' not in l:
 l = self.insequence.readline()
 if not l:
 raise StopIteration
 return l.replace('\tWARNING', '')

with open(inname) as infile:
    with open(outname, "w") as outfile:
        filter = WarningFilter(infile)
        for l in filter:
            outfile.write(l)
```

毫无疑问：它是如此丑陋且难以阅读，以至于您甚至无法分辨正在发生的事情。 我们创建了一个将文件对象作为输入的对象，并提供了像任何迭代器一样的`__next__`方法。

此`__next__`方法从文件中读取行，如果不是`WARNING`行，则将其丢弃。 当遇到`WARNING`行时，将其返回。 然后`for`循环将再次调用`__next__`以处理下一个`WARNING`行。 当我们的行数用完时，我们会举起`StopIteration`告诉循环我们已经完成了迭代。 与其他示例相比，它很丑陋，但功能也很强大。 现在我们有了一堂课，我们可以用它做任何想要的事情。

有了这些背景知识，我们终于可以看到发电机在起作用。 下一个示例将*与上一个完全相同*：它使用`__next__`方法创建一个对象，当该对象不在输入中时，该对象将引发`StopIteration`：

```py
import sys
inname, outname = sys.argv[1:3]

def warnings_filter(insequence):
 for l in insequence:
 if 'WARNING' in l:
 yield l.replace('\tWARNING', '')

with open(inname) as infile:
    with open(outname, "w") as outfile:
        filter = warnings_filter(infile)
        for l in filter:
            outfile.write(l)
```

好的，这很容易理解，也许……至少很短。 但是这里到底发生了什么，这毫无意义。 什么是`yield`？

实际上，`yield`是生成器的密钥。 当 Python 在函数中看到`yield`时，它将采用该函数并将其包装在一个对象中，这与前面的示例不同。 认为`yield`语句类似于`return`语句； 它退出函数并返回一行。 但是，与`return`不同的是，当再次调用该函数（通过`next()`）时，它将在它停下的位置（在`yield`语句之后的行上）而不是在函数的开始处开始。 在此示例中，在`yield`语句之后没有行，因此它跳转到`for`循环的下一个迭代。 由于`yield`语句位于`if`语句内部，因此仅产生包含`WARNING`的行。

看起来这只是一个循环遍历的函数，它实际上是在创建一种特殊类型的对象，即生成器对象：

```py
>>> print(warnings_filter([]))
<generator object warnings_filter at 0xb728c6bc>

```

我将一个空列表传递给该函数以充当迭代器。 该函数所做的全部工作就是创建并返回一个生成器对象。 该对象上具有`__iter__`和`__next__`方法，就像我们在上一个示例中创建的那样。 每当调用`__next__`时，生成器都会运行该函数，直到找到`yield`语句。 然后，它从`yield`返回值，下次调用`__next__`时，它将从中断处开始取值。

生成器的这种用法并不先进，但是如果您不知道该函数正在创建对象，那么它看起来就像魔术。 这个例子很简单，但是您可以通过在单个函数中多次调用`yield`来获得真正强大的效果。 生成器将仅在最近的`yield`处接收并继续到下一个。

## 从另一个可迭代项目获得收益

通常，当我们构建一个生成器函数时，我们最终会遇到以下情况：要从另一个可迭代对象中获取数据，可能是我们在生成器内部构造的列表理解或生成器表达式，或者可能是一些传递的外部项 进入功能。 通过循环遍历可迭代项并逐个生成每个项，这始终是可能的。 但是，在 Python 3.3 版中，Python 开发人员引入了新的语法，以使其更加优雅。

让我们稍微修改一下生成器示例，以便它接受文件名而不是接受行序列。 当将对象与特定范例联系在一起时，通常会对此感到不满意。 在可能的情况下，我们应该对迭代器进行操作作为输入。 这样，无论日志行来自文件，内存还是基于 Web 的日志聚合器，都可以使用相同的功能。 因此，出于教学原因，设计了以下示例。

此版本的代码说明，生成器可以在从另一个可迭代项（在这种情况下为生成器表达式）产生信息之前进行一些基本设置：

```py
import sys
inname, outname = sys.argv[1:3]

def warnings_filter(infilename):
    with open(infilename) as infile:
 yield from (
 l.replace('\tWARNING', '')
 for l in infile
 if 'WARNING' in l
 )

filter = warnings_filter(inname)
with open(outname, "w") as outfile:
    for l in filter:
        outfile.write(l)
```

此代码将前面示例中的`for`循环组合到生成器表达式中。 注意，我如何将生成器表达式的三个子句（转换，循环和过滤器）放在单独的行上，以使其更具可读性。 还要注意，这种转换的作用还不够。 前面带有`for`循环的示例更具可读性。

因此，让我们考虑一个比其他示例更具可读性的示例。 构造一个可以从多个其他生成器产生数据的生成器可能很有用。 例如，`itertools.chain`函数从可迭代对象开始按顺序生成数据，直到它们全部用尽为止。 使用`yield from`语法可能太容易实现了，所以让我们考虑一个经典的计算机科学问题：走一棵普通的树。

通用树数据结构的常见实现是计算机的文件系统。 让我们在 Unix 文件系统中为一些文件夹和文件建模，以便我们可以使用`yield from`有效地遍历它们：

```py
class File:
    def __init__(self, name):
        self.name = name

class Folder(File):
    def __init__(self, name):
        super().__init__(name)
        self.children = []

root = Folder('')
etc = Folder('etc')
root.children.append(etc)
etc.children.append(File('passwd'))
etc.children.append(File('groups'))
httpd = Folder('httpd')
etc.children.append(httpd)
httpd.children.append(File('http.conf'))
var = Folder('var')
root.children.append(var)
log = Folder('log')
var.children.append(log)
log.children.append(File('messages'))
log.children.append(File('kernel'))
```

这个设置代码看起来需要很多工作，但是在实际的文件系统中，它将涉及更多的工作。 我们必须从硬盘读取数据并将其结构化为树。 然而，一旦进入内存，输出文件系统中每个文件的代码就会非常优雅：

```py
def walk(file):
    if isinstance(file, Folder):
 yield file.name + '/'
        for f in file.children:
 yield from walk(f)
    else:
 yield file.name

```

如果此代码遇到目录，它将递归地要求`walk()`生成从属于其每个子级的所有文件的列表，然后生成所有这些数据以及其自己的文件名。 在遇到普通文件的简单情况下，它只是产生该名称。

顺便说一句，在不使用生成器的情况下解决上述问题非常棘手，以至于这个问题是常见的面试问题。 如果您按如下所示回答，请准备好让您的面试官印象深刻并有些恼怒，以至于您这么容易回答。 他们可能会要求您确切说明正在发生的事情。 当然，有了本章所遵循的原则，您将不会有任何问题。

在编写链接生成器时，`yield from`语法是的有用快捷方式，但更通常用于其他目的：通过协程传递数据。 我们将在第 13 章，“并发”中看到许多这样的示例，但是现在让我们发现协程是什么。

# 协程

协程是非常强大的构造，通常与生成器混淆。 许多作者不恰当地将协程描述为“带有一些额外语法的生成器”。 这是一个容易犯的错误，就像在 Python 2.5 中那样，当引入协程时，它们被表示为“我们在生成器语法中添加了`send`方法”。 当您在 Python 中创建协程时，返回的对象是生成器，这一事实使情况更加复杂。 实际上，差异更加细微了，在您看到一些示例后，这将变得更加有意义。

### 注意

虽然 Python 中的协程目前与生成器语法紧密相关，但它们只是与我们一直在讨论的迭代器协议表面上相关。 即将发布的 Python 3.5 发行版使协同程序成为一个真正的独立对象，并将提供与之协同工作的新语法。

要记住的另一件事是，协程很难理解。 它们并没有在野外经常使用，您可能会跳过本节而愉快地使用 Python 开发多年，而不会丢失甚至遇到它们。 有几个库广泛使用协程（主要用于并发或异步编程），但是通常编写它们的目的是使您可以在不真正了解协程工作的情况下使用协程！ 因此，如果您在本节中迷路，请不要绝望。

但是，通过研究以下示例，您不会迷路。 这是最简单的协程之一。 它使我们能够保持可以通过任意值增加的运行记录：

```py
def tally():
    score = 0
    while True:
 increment = yield score
        score += increment
```

这段代码看起来像黑魔法，可能行不通，因此在逐行描述之前，我们将看到它起作用。 这个简单的对象可以由棒球队的计分应用使用。 每个团队都可以保留单独的得分，并且每半局结束时，他们的得分可以增加累积的跑步次数。 看一下这个互动环节：

```py
>>> white_sox = tally()
>>> blue_jays = tally()
>>> next(white_sox)
0
>>> next(blue_jays)
0
>>> white_sox.send(3)
3
>>> blue_jays.send(2)
2
>>> white_sox.send(2)
5
>>> blue_jays.send(4)
6

```

首先，我们构造两个`tally`对象，每个团队一个。 是的，它们看起来像函数，但是和上一节中的生成器对象一样，函数内部有`yield`语句这一事实告诉 Python 付出了很大的努力才能将简单的函数转换为对象。

然后，我们在每个协程对象上调用`next()`。 这与在任何生成器上调用 next 相同，也就是说，它将执行每一行代码，直到遇到`yield`语句，然后返回该点的值，然后*暂停*直到 下一个`next()`通话。

到目前为止，没有什么新的。 但是，请回顾一下协程中的`yield`语句：

```py
increment = yield score
```

与生成器不同，此 yield 函数看起来应该返回值并将其分配给变量。 实际上，这就是正在发生的事情。 协程仍在`yield`语句处暂停，并等待再次调用`next()`再次激活。

或者更确切地说，正如您在交互式会话中看到的一样，是调用名为`send()`的方法。 除了将生成器前进到下一个`yield`语句之外，`send()`方法与`next()`的作用完全相同*。 它还允许您从生成器外部传递值。 此值分配给`yield`语句的左侧。*

对于许多人来说，真正令人困惑的是发生这种情况的顺序：

*   `yield`发生并且发生器暂停
*   `send()`从功能外部出现，发生器唤醒
*   发送的值分配给`yield`语句的左侧
*   生成器继续处理，直到遇到另一个`yield`语句为止

因此，在此特定示例中，在构造协程并通过调用`next()`将其推进到`yield`语句之后，每次对`send()`的后续调用都将一个值传递到协程中，该值将该值添加到其分数中 ，返回到`while`循环的顶部，并继续进行处理，直到达到`yield`语句为止。 `yield`语句返回一个值，该值成为对`send`的最新调用的返回值。 不要错过：`send()`方法不仅向生成器提交值，还像`next()`一样从即将到来的`yield`语句中返回值。 这就是我们定义生成器和协程之间的区别的方式：生成器仅产生值，而协程也可以消耗它们。

### 注意

`next(i)`，`i.__next__()`和`i.send(value)`的行为和语法非常不直观且令人沮丧。 第一个是普通函数，第二个是特殊方法，最后一个是普通方法。 但是这三个都做同样的事情：使生成器前进，直到产生一个值并暂停。 此外，可以通过调用`i.send(None)`复制`next()`函数和相关方法。 在这里拥有两个不同的方法名称是有价值的，因为它可以帮助我们的代码阅读者轻松查看它们是与协程还是生成器进行交互。 我只是发现以下事实：在一种情况下，这是一个函数调用，而在另一种情况下，这是一种普通的方法，有些恼人。

## 返回日志解析

当然，可以使用几个整数变量并在其上调用`x += increment`轻松地对前面的示例进行编码。 让我们看一下第二个示例，其中协程实际上为我们节省了一些代码。 这个示例是我在实际工作中必须解决的问题的简化版本（出于教学原因）。 它在逻辑上是从前面有关处理日志文件的讨论中得出的，这完全是偶然的； 这些示例是为该模块的第一版编写的，而四年后出现了这个问题！

Linux 内核日志包含一些看上去有些象但并不完全象样的行，这与下面的行不同：

```py
unrelated log messages
sd 0:0:0:0 Attached Disk Drive
unrelated log messages
sd 0:0:0:0 (SERIAL=ZZ12345)
unrelated log messages
sd 0:0:0:0 [sda] Options
unrelated log messages
XFS ERROR [sda]
unrelated log messages
sd 2:0:0:1 Attached Disk Drive
unrelated log messages
sd 2:0:0:1 (SERIAL=ZZ67890)
unrelated log messages
sd 2:0:0:1 [sdb] Options
unrelated log messages
sd 3:0:1:8 Attached Disk Drive
unrelated log messages
sd 3:0:1:8 (SERIAL=WW11111)
unrelated log messages
sd 3:0:1:8 [sdc] Options
unrelated log messages
XFS ERROR [sdc]
unrelated log messages
```

有一堆散布的内核日志消息，其中一些消息与硬盘有关。 硬盘消息可能会散布在其他消息中，但是它们以可预测的格式和顺序出现，其中具有已知序列号的特定驱动器与总线标识符（例如`0:0:0:0`）和块设备标识符相关联 （例如`sda`）与该总线关联。 最后，如果驱动器的文件系统损坏，它可能会因 XFS 错误而失败。

现在，在给定前面的日志文件的情况下，我们需要解决的问题是如何获取所有带有 XFS 错误的驱动器的序列号。 数据中心技术人员以后可能会使用该序列号来标识和更换驱动器。

我们知道可以使用正则表达式来识别各个行，但是我们在循环遍历这些行时必须更改正则表达式，因为我们将根据之前的发现寻找不同的东西。 另一个困难的一点是，如果我们找到一个错误字符串，则有关哪条总线包含该字符串以及该总线上的驱动器附带的序列号的信息已得到处理。 通过以相反的顺序遍历文件的行可以轻松解决此问题。

在查看此示例之前，请注意—基于协程的解决方案所需的代码量很少：

```py
import re

def match_regex(filename, regex):
    with open(filename) as file:
        lines = file.readlines()
    for line in reversed(lines):
        match = re.match(regex, line)
        if match:
 regex = yield match.groups()[0]

def get_serials(filename):
    ERROR_RE = 'XFS ERROR (\[sd[a-z]\])'
 matcher = match_regex(filename, ERROR_RE)
    device = next(matcher)
    while True:
        bus = matcher.send(
            '(sd \S+) {}.*'.format(re.escape(device)))
        serial = matcher.send('{} \(SERIAL=([^)]*)\)'.format(bus))
 yield serial
        device = matcher.send(ERROR_RE)

for serial_number in get_serials('EXAMPLE_LOG.log'):
    print(serial_number)
```

此代码将作业整齐地分为两个单独的任务。 第一个任务是遍历所有行并吐出与给定正则表达式匹配的所有行。 第二项任务是与第一个任务进行交互，并提供有关在任何给定时间应该搜索哪个正则表达式的指导。

首先查看`match_regex`协程。 请记住，它在构造时不会执行任何代码； 相反，它只是创建一个协程对象。 一旦构建，协程外部的某个人最终将调用`next()`来开始运行代码，这时它将存储两个变量`filename`和`regex`的状态。 然后，它读取文件中的所有行，并以相反的顺序遍历它们。 将每一行与传入的正则表达式进行比较，直到找到匹配项。 找到匹配项后，协程将从正则表达式中产生第一组并等待。

在将来的某些点，其他代码将发送一个新的正则表达式进行搜索。 注意协程从不关心它要匹配的正则表达式。 它只是遍历行并将它们与正则表达式进行比较。 决定要提供什么正则表达式是其他人的责任。

在这种情况下，其他人是`get_serials`生成器。 它不在乎文件中的行，实际上它甚至不知道它们。 它要做的第一件事是从`match_regex`协程构造函数创建一个`matcher`对象，并为其提供默认的正则表达式进行搜索。 它将协程前进到其第一个`yield`，并存储它返回的值。 然后进入一个循环，该循环指示匹配器对象根据所存储的设备 ID 搜索总线 ID，然后根据该总线 ID 搜索序列号。

在指示匹配器找到另一个设备 ID 并重复该循环之前，它会闲置地将序列号输出到`for`外部。

基本上，协程（`match_regex`，因为它使用`regex = yield`语法）作业是在文件中搜索下一个重要行，而生成器的[`get_serial`，它使用`yield`语法而没有分配）。 决定哪条线很重要。 生成器具有有关此特定问题的信息，例如文件中将显示哪些订单行。 另一方面，协程可以插入任何需要在文件中搜索给定正则表达式的问题。

## 关闭协程并引发异常

普通生成器通过升高`StopIteration`发出信号从内部退出。 如果我们将多个生成器链接在一起（例如，通过从另一个生成器内部迭代一个生成器），则`StopIteration`异常将向外传播。 最终，它将到达`for`循环，该循环将看到异常并知道该退出循环了。

协程通常不遵循迭代机制。 通常不将数据拉入直到遇到异常，而是通常将数据压入其中（使用`send`）。 进行推送的实体通常是负责告知协程完成时间的实体。 它通过在相关协程上调用`close()`方法来实现。

调用时，`close()`方法会在协程为等待发送值时引发`GeneratorExit`异常。协程通常将其`yield`语句包装在`try` ... `finally`块，以便可以执行任何清理任务（例如关闭关联的文件或套接字）。

如果需要在协程内部引发异常，则可以类似的方式使用`throw()`方法。 它接受带有可选`value`和`traceback`参数的异常类型。 当我们在一个协程中遇到异常并希望在维持回溯的同时在相邻协程中发生异常时，后者非常有用。

如果您要构建强大的基于协程的库，那么这两个功能都是至关重要的，但是在日常编码工作中我们不太可能遇到它们。

## 协程，生成器和函数之间的关系

我们已经看到了协程的实际作用，所以现在让我们回到有关它们与生成器的关系的讨论。 在 Python 中（通常如此），区别非常模糊。 实际上，所有协程都是生成器对象，并且作者经常将这两个术语互换使用。 有时，他们将协程描述为生成器的子集（只有从 yield 返回值的生成器才被视为协程）。 正如我们在上一节中所见，在技术上，这在 Python 中是正确的。

但是，在理论计算机科学的更大范围内，协程被认为是更通用的原理，而生成器是协程的一种特定类型。 此外，正常功能是协程的另一个不同子集。

协程是一种例程，可以使数据在一个或多个点处传入并在一个或多个点处将其取出。 在 Python 中，数据传入和传出的点是`yield`语句。

函数或子例程是协程的最简单类型。 您可以在某一点传递数据，而在函数返回时在另一点获取数据。 一个函数可以有多个`return`语句，但是对于该函数的任何给定调用，只能调用其中一个。

最后，生成器是一种协程，可以在某一点传递数据，但可以在多个点传递数据。 在 Python 中，数据将在`yield`语句中传递出去，但是您不能传递回数据。如果调用`send`，则数据将被静默丢弃。

因此，从理论上讲，生成器是协程的类型，函数是协程的类型，并且有些协程既不是函数也不是生成器。 这很简单，是吗？ 那么为什么在 Python 中感觉更复杂？

在 Python 中，生成器和协程都是使用类似于我们正在构造函数的语法构造的。 但是生成的对象根本不是一个函数。 这是完全不同的对象。 当然，功能也是对象。 但是它们具有不同的界面。 函数是可调用的并返回值，生成器使用`next()`提取数据，协程使用`send`推送数据。

# 案例研究

数据科学是 Python 在当今最流行的领域之一。 让我们实现基本的机器学习算法！ 机器学习是一个巨大的话题，但是总体思路是通过使用从过去的数据中获得的知识来对未来的数据进行预测或分类。 此类算法的用途十分广泛，数据科学家每天都在寻找新的方法来应用机器学习。 一些重要的机器学习应用包括计算机视觉（例如图像分类或面部识别），产品推荐，识别垃圾邮件和语音识别。 我们来看一个更简单的问题：给定 RGB 颜色定义，人类将这个颜色标识为什么名字？

在标准 RGB 颜色空间中，有超过 1600 万种颜色，而人类只想出了其中一小部分的名字。 尽管有成千上万个名称（有些很荒谬；可以去任何汽车经销商或化妆品店），但让我们建立一个分类器，尝试将 RGB 空间划分为基本颜色：

*   红色的
*   紫色的
*   蓝色
*   绿色的
*   黄色的
*   橘子
*   灰色的
*   白色的
*   粉色的

我们需要的第一件事是训练我们的算法的数据集。 在生产系统中，您可能会刮擦*颜色*网站列表或调查数千人。 取而代之的是，我创建了一个简单应用，该应用呈现随机颜色并要求用户选择上述九个选项之一来对其进行分类。 该应用包含在`kivy_color_classifier`目录中的本章示例代码中，但是我们将不讨论此代码的细节，因为它的唯一目的是生成示例数据。

### 注意

Kivy 有一个设计精良的面向对象的 API，您可能需要自己研究一下。 如果您想开发在许多系统上运行的图形程序，从笔记本电脑到手机，您可能需要查看我的模块*在 Kivy 中创建应用*，*O'Reilly*。

就本案例研究而言，关于该应用的重要事项是输出，它是**逗号分隔值**（**CSV**）文件，每行包含四个值：红色 ，绿色和蓝色值（表示为介于零和一之间的浮点数），以及用户为该颜色分配的前九个名称之一。 数据集看起来像这样：

```py
0.30928279150905513,0.7536768153744394,0.3244011790604804,Green
0.4991001855115986,0.6394567277907686,0.6340502030888825,Grey
0.21132621004927998,0.3307376167520666,0.704037576789711,Blue
0.7260420945787928,0.4025279573860123,0.49781705131696363,Pink
0.706469868610228,0.28530423638868196,0.7880240251003464,Purple
0.692243900051664,0.7053550777777416,0.1845069151913028,Yellow
0.3628979381122397,0.11079495501215897,0.26924540840045075,Purple
0.611273677646518,0.48798521783547677,0.5346130557761224,Purple
.
.
.
0.4014121109376566,0.42176706818252674,0.9601866228083298,Blue
0.17750449496124632,0.8008214961070862,0.5073944321437429,Green
```

在感到无聊之前，我做了 200 个数据点（其中有很多是不真实的），并决定是时候对该数据集进行机器学习了。 如果您想使用我的数据，这些数据点将随本章的示例一起提供（没有人告诉我我是色盲的，因此应该是合理的）。

我们将实现一种更简单的机器学习算法，称为 k 最近邻居。 该算法依赖于数据集中各点之间的某种“距离”计算（在我们的示例中，我们可以使用勾股定理的三维版本）。 给定一个新的数据点，它会找到一定数量的数据点（在 k 个最近的邻居中称为 k），该数据点在通过该距离计算进行测量时最接近它。 然后，它以某种方式组合这些数据点（平均值可能适用于线性计算；对于我们的分类问题，我们将使用该模式），并返回结果。

我们不会过多地介绍算法的作用。 相反，我们将重点介绍一些可以将迭代器模式或迭代器协议应用于此问题的方法。

现在让我们编写一个程序，该程序按顺序执行以下步骤：

1.  从文件中加载样本数据并从中构建模型。
2.  生成 100 种随机颜色。
3.  对每种颜色进行分类，然后以与输入相同的格式将其输出到文件中。

有了第二个 CSV 文件后，另一个 Kivy 程序可以加载该文件并渲染每种颜色，要求人类用户确认或否认预测的准确性，从而告知我们算法和初始数据集的准确性如何。 。

第一步是一个相当简单的生成器，用于加载 CSV 数据并将其转换为适合我们需求的格式：

```py
import csv

dataset_filename = 'colors.csv'

def load_colors(filename):
    with open(filename) as dataset_file:
 lines = csv.reader(dataset_file)
        for line in lines:
 yield tuple(float(y) for y in line[0:3]), line[3]

```

我们之前没有见过`csv.reader`功能。 它在文件中的各行上返回一个迭代器。 迭代器返回的每个值都是一个字符串列表。 在我们的情况下，我们可能会用逗号分开并且很好，但是`csv.reader`还负责管理引号和逗号分隔的值格式的各种其他细微差别。

然后，我们遍历这些行并将其转换为颜色和名称的元组，其中颜色是三个浮点值整数的元组。 该元组是使用生成器表达式构造的。 可能有更多可读的方法来构造该元组。 您是否认为代码简短和生成器表达式的速度值得混淆？ 它没有返回颜色元组列表，而是一次生成一个颜色元组，从而构造了一个生成器对象。

现在，我们需要一百种随机颜色。 有很多方法可以做到这一点：

*   具有嵌套生成器表达式的列表推导：`[tuple(random() for r in range(3)) for r in range(100)]`
*   基本的生成器功能
*   实现`__iter__`和`__next__`协议的类
*   通过协程管道推送数据
*   即使是基本的`for`循环

生成器版本似乎是最易读的，因此让我们将该函数添加到程序中：

```py
from random import random

def generate_colors(count=100):
    for i in range(count):
 yield (random(), random(), random())

```

注意我们如何参数化要生成的颜色数量。 现在，我们可以将来将此功能重新用于其他颜色生成任务。

现在，在进行分类步骤之前，我们需要一个函数来计算两种颜色之间的“距离”。 由于可以将颜色视为三维（例如，红色，绿色和蓝色可以映射到`x`，`y`和`z`轴） ，让我们使用一些基本的数学运算：

```py
import math

def color_distance(color1, color2):
    channels = zip(color1, color2)
    sum_distance_squared = 0
    for c1, c2 in channels:
        sum_distance_squared += (c1 - c2) ** 2
    return math.sqrt(sum_distance_squared)
```

这是一个非常基本的功能； 它看起来甚至没有使用迭代器协议。 没有`yield`功能，也没有理解力。 但是，有一个`for`循环，对`zip`函数的调用也正在执行一些实际的迭代（请记住`zip`从每个输入迭代器中生成包含一个元素的元组）。

但是请注意，在我们的 k 近邻算法中，该函数将被调用很多次。 如果我们的代码运行太慢，并且我们能够将该函数识别为瓶颈，则我们可能希望将其替换为可读性较低但更优化的生成器表达式：

```py
def color_distance(color1, color2):
    return math.sqrt(sum((x[0] - x[1]) ** 2 for x in zip(
    color1, color2)))
```

但是，强烈建议您在证明可读版本太慢之前不要进行此类优化。

现在我们已经有了一些管道，让我们进行实际的 k 最近邻居实现。 这似乎是使用协程的好地方。 这里带有一些测试代码，以确保产生合理的值：

```py
def nearest_neighbors(model_colors, num_neighbors):
    model = list(model_colors)
 target = yield
    while True:
        distances = sorted(
            ((color_distance(c[0], target), c) for c in model),
        )
 target = yield [
 d[1] for d in distances[0:num_neighbors]
 ]

model_colors = load_colors(dataset_filename)
target_colors = generate_colors(3)
get_neighbors = nearest_neighbors(model_colors, 5)
next(get_neighbors)

for color in target_colors:
    distances = get_neighbors.send(color)
    print(color)
    for d in distances:
        print(color_distance(color, d[0]), d[1])
```

协程接受两个参数，即用作模型的颜色列表和要查询的邻居数。 它将模型转换为列表，因为它将多次迭代。 在协程的主体中，它使用`yield`语法接受 RGB 颜色值的元组。 然后，它将对`sorted`的调用与奇数生成器表达式组合在一起。 看看是否可以弄清楚该生成器表达式在做什么。

它为模型中的每种颜色返回一个元组`(distance, color_data)`。 请记住，模型本身包含`(color, name)`的元组，其中`color`是三个 RGB 值的元组。 因此，生成器将在看起来像这样的怪异数据结构上返回迭代器：

```py
(distance, (r, g, b), color_name)
```

然后`sorted`调用按结果的第一个元素（即距离）对结果进行排序。 这是一段复杂的代码，根本不是面向对象的。 您可能需要将其分解为正常的`for`循环，以确保您了解生成器表达式在做什么。 想像一下如果要将键参数传递给已排序的函数而不是构造元组，则这段代码的外观可能是一个很好的练习。

`yield`语句稍微复杂一些； 它从前 k 个`(distance, color_data)`元组的每一个中提取第二个值。 更具体地说，它为具有最小距离的 k 值生成`((r, g, b), color_name)`元组。 或者，如果您更喜欢抽象术语，它会在给定模型中得出目标的 k 最近邻。

剩下的代码只是测试这种方法的样板。 它构建模型和颜色生成器，启动协程，并在`for`循环中打印结果。

剩下的两个任务是根据最近的邻居选择一种颜色，并将结果输出到 CSV 文件。 让我们再增加两个协程来完成这些任务。 让我们先进行输出，因为它可以独立测试：

```py
def write_results(filename="output.csv"):
    with open(filename, "w") as file:
        writer = csv.writer(file)
        while True:
 color, name = yield
            writer.writerow(list(color) + [name])

results = write_results()
next(results)
for i in range(3):
    print(i)
    results.send(((i, i, i), i * 10))
```

该协程将打开的文件保持为状态，并在使用`send()`发送时向其中写入代码行。 测试代码确保协程正常工作，因此现在我们可以将两个协程与第三个协程连接起来。

第二个协程使用了一些奇怪的技巧：

```py
from collections import Counter
def name_colors(get_neighbors):
 color = yield
    while True:
 near = get_neighbors.send(color)
        name_guess = Counter(
            n[1] for n in near).most_common(1)[0][0]
 color = yield name_guess

```

该协程将现有协程接受*作为其参数*。 在这种情况下，它是`nearest_neighbors`的实例。 这段代码基本上代理了通过该`nearest_neighbors`实例发送给它的所有值。 然后，它将对结果进行一些处理，以从返回的值中获得最普通的颜色。 在这种情况下，修改原始协程以返回名称可能同样有意义，因为它没有被用于其他任何用途。 但是，在许多情况下，传递协程是有用的。 这是我们的做法。

现在我们要做的就是将这些协程和管道连接在一起，并通过一个函数调用启动该过程：

```py
def process_colors(dataset_filename="colors.csv"):
    model_colors = load_colors(dataset_filename)
    get_neighbors = nearest_neighbors(model_colors, 5)
 get_color_name = name_colors(get_neighbors)
    output = write_results()
 next(output)
 next(get_neighbors)
 next(get_color_name)

    for color in generate_colors():
 name = get_color_name.send(color)
 output.send((color, name))

process_colors()
```

因此，与我们定义的几乎所有其他函数不同，该函数是完全正常的函数，没有任何`yield`语句。 它不会变成协程或生成器对象。 但是，它确实构造了一个生成器和三个协程。 注意`get_neighbors`协程如何传递到`name_colors`的构造函数中？ 请注意通过调用`next`将所有三个协程推进到其第一个`yield`语句。

创建完所有管道后，我们使用`for`循环将生成的每种颜色发送到`get_color_name`协程，然后将该协程产生的每个值通过管道传递到输出协程，然后将其写入到 文件。

就是这样！ 我创建了第二个 Kivy 应用，该应用加载生成的 CSV 文件并将颜色呈现给用户。 用户可以选择*是*或*否*，具体取决于他们是否认为机器学习算法做出的选择与他们会做出的选择相匹配。 这在科学上是不准确的（对于观察偏差已经成熟），但是足够玩了。 用我的眼睛，它成功了大约 84％，这比我 12 年级的平均水平要好。 对于我们的第一个机器学习体验来说还不错，是吗？

您可能想知道，“这与面向对象的编程有什么关系？此代码中甚至没有一个类！”。 在某些方面，您将是对的； 协程和生成器通常都不被认为是面向对象的。 但是，创建它们的函数将返回对象。 实际上，您可以将这些函数视为构造函数。 构造的对象具有适当的`send()`和`__next__()`方法。 基本上，协程/生成器语法是特定对象的语法快捷方式，如果没有该对象，创建起来将非常冗长。

此案例研究是自下而上设计的练习。 我们创建了执行特定任务的各种低级对象，并将它们最终挂在一起。 我发现使用协程开发时这是一种常见的做法。 自上而下的替代设计有时会导致更多的整体代码段，而不是唯一的单独代码段。 总的来说，我们希望在太大的方法和太小的方法之间找到一个快乐的中介，很难看到它们如何组合在一起。 当然，无论是否像在此那样使用迭代器协议，都是如此。

![Case study](img/C_01_10.jpg)

![Case study](img/Summary_01_10.jpg)

![Case study](img/Yourprogress_01.jpg)